from collections.abc import Sequence

from supriya import utils

from .bases import UGen, param, ugen


@ugen(ar=True, kr=True, is_input=True, is_multichannel=True)
class In(UGen):
    """
    A bus input unit generator.

    ::

        >>> supriya.ugens.In.ar(bus=0, channel_count=4)
        UGenArray({4})
    """

    bus = param(0.0)


@ugen(ar=True, kr=True, is_input=True, is_multichannel=True)
class InFeedback(UGen):
    """
    A bus input unit generator.

    Reads signal from a bus with a current or one cycle old timestamp.

    ::

        >>> in_feedback = supriya.ugens.InFeedback.ar(
        ...     bus=0,
        ...     channel_count=2,
        ... )
        >>> in_feedback
        UGenArray({2})
    """

    bus = param(0.0)


@ugen(ar=True, kr=True, is_multichannel=True)
class LocalIn(UGen):
    """
    A SynthDef-local bus input.

    ::

        >>> supriya.ugens.LocalIn.ar(channel_count=2)
        UGenArray({2})
    """

    ### CLASS VARIABLES ###

    default = param(0.0, unexpanded=True)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, channel_count=1, default=0):
        self._channel_count = int(channel_count)
        if not isinstance(default, Sequence):
            default = (default,)
        default = (float(_) for _ in default)
        default = utils.repeat_sequence_to_length(default, channel_count)
        default = list(default)[:channel_count]
        UGen.__init__(self, calculation_rate=calculation_rate, default=default)


@ugen(ar=True, kr=True, channel_count=0, fixed_channel_count=True)
class LocalOut(UGen):
    """
    A SynthDef-local bus output.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.LocalOut.ar(
        ...     source=source,
        ... )
        LocalOut.ar()
    """

    source = param(None, unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class OffsetOut(UGen):
    """
    A bus output unit generator with sample-accurate timing.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.OffsetOut.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        OffsetOut.ar()
    """

    bus = param(0)
    source = param(None, unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class Out(UGen):
    """
    A bus output unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.Out.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        Out.ar()
    """

    bus = param(0)
    source = param(None, unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class ReplaceOut(UGen):
    """
    An overwriting bus output unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.ReplaceOut.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        ReplaceOut.ar()
    """

    bus = param(0)
    source = param(None, unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class XOut(UGen):
    """
    A cross-fading bus output unit generator.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> xout = supriya.ugens.XOut.ar(
        ...     bus=0,
        ...     crossfade=0.5,
        ...     source=source,
        ... )
        >>> xout
        XOut.ar()
    """

    bus = param(0)
    crossfade = param(0.0)
    source = param(None, unexpanded=True)



import asyncio
import atexit
import enum
import logging
import os
import platform
import signal
import subprocess
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import uqbar.io
import uqbar.objects

from .exceptions import ServerCannotBoot

logger = logging.getLogger(__name__)

DEFAULT_IP_ADDRESS = "127.0.0.1"
DEFAULT_PORT = 57110
ENVAR_SERVER_EXECUTABLE = "SUPRIYA_SERVER_EXECUTABLE"


@dataclass(frozen=True)
class Options:
    """
    SuperCollider server options configuration.
    """

    ### CLASS VARIABLES ###

    audio_bus_channel_count: int = 1024
    block_size: int = 64
    buffer_count: int = 1024
    control_bus_channel_count: int = 16384
    executable: Optional[str] = None
    hardware_buffer_size: Optional[int] = None
    initial_node_id: int = 1000
    input_bus_channel_count: int = 8
    input_device: Optional[str] = None
    input_stream_mask: str = ""
    ip_address: str = DEFAULT_IP_ADDRESS
    load_synthdefs: bool = True
    maximum_logins: int = 1
    maximum_node_count: int = 1024
    maximum_synthdef_count: int = 1024
    memory_locking: bool = False
    memory_size: int = 8192
    output_bus_channel_count: int = 8
    output_device: Optional[str] = None
    output_stream_mask: str = ""
    password: Optional[str] = None
    port: int = DEFAULT_PORT
    protocol: str = "udp"
    random_number_generator_count: int = 64
    remote_control_volume: bool = False
    realtime: bool = True
    restricted_path: Optional[str] = None
    sample_rate: Optional[int] = None
    threads: Optional[int] = None
    ugen_plugins_path: Optional[str] = None
    verbosity: int = 0
    wire_buffer_count: int = 64
    zero_configuration: bool = False

    ### INITIALIZER ###

    def __post_init__(self):
        if self.input_bus_channel_count is None:
            object.__setattr__(self, "input_bus_channel_count", 8)
        if self.output_bus_channel_count is None:
            object.__setattr__(self, "output_bus_channel_count", 8)
        if self.input_bus_channel_count < 0:
            raise ValueError(self.input_bus_channel_count)
        if self.output_bus_channel_count < 0:
            raise ValueError(self.output_bus_channel_count)
        if self.audio_bus_channel_count < (
            self.input_bus_channel_count + self.output_bus_channel_count
        ):
            raise ValueError("Insufficient audio buses")

    ### CLASS VARIABLES ###

    def __repr__(self):
        return uqbar.objects.get_repr(self, multiline=True, suppress_defaults=False)

    def __iter__(self):
        return (arg for arg in self.serialize())

    ### PUBLIC METHODS ###

    def get_audio_bus_ids(self, client_id: int) -> Tuple[int, int]:
        audio_buses_per_client = (
            self.private_audio_bus_channel_count // self.maximum_logins
        )
        minimum = self.first_private_bus_id + (client_id * audio_buses_per_client)
        maximum = self.first_private_bus_id + ((client_id + 1) * audio_buses_per_client)
        return minimum, maximum

    def get_buffer_ids(self, client_id: int) -> Tuple[int, int]:
        buffers_per_client = self.buffer_count // self.maximum_logins
        minimum = client_id * buffers_per_client
        maximum = (client_id + 1) * buffers_per_client
        return minimum, maximum

    def get_control_bus_ids(self, client_id: int) -> Tuple[int, int]:
        control_buses_per_client = self.control_bus_channel_count // self.maximum_logins
        minimum = client_id * control_buses_per_client
        maximum = (client_id + 1) * control_buses_per_client
        return minimum, maximum

    def get_sync_ids(self, client_id: int) -> Tuple[int, int]:
        return client_id << 26, (client_id + 1) << 26

    def serialize(self) -> List[str]:
        result = [str(find(self.executable))]
        pairs: Dict[str, Optional[str]] = {}
        if self.realtime:
            if self.protocol == "tcp":
                pairs["-t"] = str(self.port)
            else:
                pairs["-u"] = str(self.port)
            if self.input_device:
                pairs["-H"] = str(self.input_device)
                if self.output_device != self.input_device:
                    result.append(str(self.output_device))
            if self.maximum_logins != 64:
                pairs["-l"] = str(self.maximum_logins)
            if self.password:
                pairs["-p"] = str(self.password)
            if self.sample_rate is not None:
                pairs["-S"] = str(self.sample_rate)
            if not self.zero_configuration:
                pairs["-R"] = "0"
        if self.audio_bus_channel_count != 1024:
            pairs["-a"] = str(self.audio_bus_channel_count)
        if self.control_bus_channel_count != 16384:
            pairs["-c"] = str(self.control_bus_channel_count)
        if self.input_bus_channel_count != 8:
            pairs["-i"] = str(self.input_bus_channel_count)
        if self.output_bus_channel_count != 8:
            pairs["-o"] = str(self.output_bus_channel_count)
        if self.buffer_count != 1024:
            pairs["-b"] = str(self.buffer_count)
        if self.maximum_node_count != 1024:
            pairs["-n"] = str(self.maximum_node_count)
        if self.maximum_synthdef_count != 1024:
            pairs["-d"] = str(self.maximum_synthdef_count)
        if self.block_size != 64:
            pairs["-z"] = str(self.block_size)
        if self.hardware_buffer_size is not None:
            pairs["-Z"] = str(self.hardware_buffer_size)
        if self.memory_size != 8192:
            pairs["-m"] = str(self.memory_size)
        if self.random_number_generator_count != 64:
            pairs["-r"] = str(self.random_number_generator_count)
        if self.wire_buffer_count != 64:
            pairs["-w"] = str(self.wire_buffer_count)
        if not self.load_synthdefs:
            pairs["-D"] = "0"
        if self.input_stream_mask:
            pairs["-I"] = str(self.input_stream_mask)
        if self.output_stream_mask:
            pairs["-O"] = str(self.output_stream_mask)
        if 0 < self.verbosity:
            pairs["-v"] = str(self.verbosity)
        if self.restricted_path is not None:
            pairs["-P"] = str(self.restricted_path)
        if self.memory_locking:
            pairs["-L"] = None
        if self.ugen_plugins_path:
            pairs["-U"] = str(self.ugen_plugins_path)
        if self.threads and find(self.executable).stem == "supernova":
            pairs["-t"] = str(self.threads)
        for key, value in sorted(pairs.items()):
            result.extend([key, value] if value is not None else [key])
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def first_private_bus_id(self):
        return self.output_bus_channel_count + self.input_bus_channel_count

    @property
    def private_audio_bus_channel_count(self):
        return (
            self.audio_bus_channel_count
            - self.input_bus_channel_count
            - self.output_bus_channel_count
        )


def find(scsynth_path=None):
    """
    Find the ``scsynth`` executable.

    The following paths, if defined, will be searched (prioritised as ordered):

    1. The absolute path ``scsynth_path``
    2. The environment variable ``SUPRIYA_SERVER_EXECUTABLE`` (pointing to the `scsynth`
       binary)
    3. The user's ``PATH``
    4. Common installation directories of the SuperCollider application.

    Returns a path to the ``scsynth`` executable. Raises ``RuntimeError`` if no path is
    found.
    """
    path = Path(scsynth_path or os.environ.get(ENVAR_SERVER_EXECUTABLE) or "scsynth")
    if path.is_absolute() and uqbar.io.find_executable(str(path)):
        return path
    path_candidates = uqbar.io.find_executable(path.name)
    if path_candidates:
        return Path(path_candidates[0])
    paths = []
    executable = scsynth_path or "scsynth"
    if Path(executable).stem == "supernova":
        executable = "supernova"
    system = platform.system()
    if system == "Linux":
        paths.extend(
            [Path("/usr/bin/" + executable), Path("/usr/local/bin/" + executable)]
        )
    elif system == "Darwin":
        paths.append(
            Path("/Applications/SuperCollider.app/Contents/Resources/" + executable)
        )
    elif system == "Windows":
        paths.extend(
            Path(r"C:\Program Files").glob(r"SuperCollider*\\" + executable + ".exe")
        )
    for path in paths:
        if path.exists():
            return path
    raise RuntimeError("Failed to locate executable")


def kill():
    with subprocess.Popen(
        ["ps", "-Af"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    ) as process:
        output = process.stdout.read()
    for line in output.decode().splitlines():
        parts = line.split()
        if not any(part in ["supernova", "scsynth"] for part in parts):
            continue
        pid = int(parts[1])
        os.kill(pid, signal.SIGKILL)


class LineStatus(enum.IntEnum):
    CONTINUE = 0
    READY = 1
    ERROR = 2


class ProcessProtocol:
    def __init__(self):
        self.is_running = False

    def boot(self, options: Options):
        raise NotImplementedError

    def quit(self):
        raise NotImplementedError

    def _handle_line(self, line):
        if line.startswith("late:"):
            logger.warning(f"Received: {line}")
        elif "error" in line.lower() or "exception" in line.lower():
            logger.error(f"Received: {line}")
        else:
            logger.info(f"Received: {line}")
        if line.startswith(("SuperCollider 3 server ready", "Supernova ready")):
            return LineStatus.READY
        elif line.startswith(("Exception", "ERROR", "*** ERROR")):
            return LineStatus.ERROR
        return LineStatus.CONTINUE


class SyncProcessProtocol(ProcessProtocol):
    def __init__(self):
        super().__init__()
        atexit.register(self.quit)

    def boot(self, options: Options):
        if self.is_running:
            return
        try:
            logger.info("Boot: {}".format(*options))
            self.process = subprocess.Popen(
                list(options),
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE,
                start_new_session=True,
            )
            start_time = time.time()
            timeout = 10
            while True:
                line = self.process.stdout.readline().decode().rstrip()  # type: ignore
                if not line:
                    continue
                line_status = self._handle_line(line)
                if line_status == LineStatus.READY:
                    break
                elif line_status == LineStatus.ERROR:
                    raise ServerCannotBoot(line)
                elif (time.time() - start_time) > timeout:
                    raise ServerCannotBoot(line)
            self.is_running = True
        except ServerCannotBoot:
            self.process.terminate()
            self.process.wait()
            raise

    def quit(self) -> None:
        if not self.is_running:
            return
        self.process.terminate()
        self.process.wait()
        self.is_running = False


class AsyncProcessProtocol(asyncio.SubprocessProtocol, ProcessProtocol):
    ### INITIALIZER ###

    def __init__(self):
        ProcessProtocol.__init__(self)
        asyncio.SubprocessProtocol.__init__(self)
        self.boot_future = asyncio.Future()
        self.exit_future = asyncio.Future()
        self.error_text = ""

    ### PUBLIC METHODS ###

    async def boot(self, options: Options):
        logger.info("Booting ...")
        if self.is_running:
            logger.info("... already booted!")
            return
        self.is_running = False
        loop = asyncio.get_running_loop()
        self.boot_future = loop.create_future()
        self.exit_future = loop.create_future()
        self.error_text = ""
        self.buffer_ = ""
        _, _ = await loop.subprocess_exec(
            lambda: self, *options, stdin=None, stderr=None
        )
        if not (await self.boot_future):
            raise ServerCannotBoot(self.error_text)

    def connection_made(self, transport):
        logger.info("Connection made!")
        self.is_running = True
        self.transport = transport

    def pipe_connection_lost(self, fd, exc):
        logger.info("Pipe connection lost!")

    def pipe_data_received(self, fd, data):
        # *nix and OSX return full lines,
        # but Windows will return partial lines
        # which obligates us to reconstruct them.
        text = self.buffer_ + data.decode().replace("\r\n", "\n")
        if "\n" in text:
            text, _, self.buffer_ = text.rpartition("\n")
            for line in text.splitlines():
                line_status = self._handle_line(line)
                if line_status == LineStatus.READY:
                    self.boot_future.set_result(True)
                    logger.info("... booted!")
                elif line_status == LineStatus.ERROR:
                    if not self.boot_future.done():
                        self.boot_future.set_result(False)
                        self.error_text = line
                    logger.info("... failed to boot!")
        else:
            self.buffer_ = text

    def process_exited(self):
        logger.info(f"Process exited with {self.transport.get_returncode()}.")
        self.is_running = False
        try:
            self.exit_future.set_result(None)
            if not self.boot_future.done():
                self.boot_future.set_result(False)
        except asyncio.exceptions.InvalidStateError:
            pass

    async def quit(self):
        logger.info("Quitting ...")
        if not self.is_running:
            logger.info("... already quit!")
            return
        self.is_running = False
        self.transport.close()
        await self.exit_future
        logger.info("... quit!")


class AsyncNonrealtimeProcessProtocol(asyncio.SubprocessProtocol):
    def __init__(self, exit_future: asyncio.Future) -> None:
        self.buffer_ = ""
        self.exit_future = exit_future

    async def run(self, command: List[str], render_directory_path: Path) -> None:
        logger.info(f"Running: {' '.join(command)}")
        _, _ = await asyncio.get_running_loop().subprocess_exec(
            lambda: self,
            *command,
            stdin=None,
            stderr=None,
            start_new_session=True,
            cwd=render_directory_path,
        )

    def handle_line(self, line: str) -> None:
        logger.debug(f"Received: {line}")

    def connection_made(self, transport):
        logger.debug("Connecting")
        self.transport = transport

    def pipe_data_received(self, fd, data):
        logger.debug(f"Data: {data}")
        # *nix and OSX return full lines,
        # but Windows will return partial lines
        # which obligates us to reconstruct them.
        text = self.buffer_ + data.decode().replace("\r\n", "\n")
        if "\n" in text:
            text, _, self.buffer_ = text.rpartition("\n")
            for line in text.splitlines():
                self.handle_line(line)
        else:
            self.buffer_ = text

    def process_exited(self):
        logger.debug(f"Exiting with {self.transport.get_returncode()}")
        self.exit_future.set_result(self.transport.get_returncode())



"""
Tools for sending, receiving and handling OSC messages.
"""

import abc
import asyncio
import collections
import contextlib
import dataclasses
import datetime
import enum
import inspect
import logging
import queue
import socket
import socketserver
import struct
import threading
import time
from collections.abc import Sequence as SequenceABC
from typing import (
    Any,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from uqbar.objects import get_repr

from .utils import group_iterable_by_count

osc_protocol_logger = logging.getLogger(__name__)
osc_in_logger = logging.getLogger("supriya.osc.in")
osc_out_logger = logging.getLogger("supriya.osc.out")
udp_in_logger = logging.getLogger("supriya.udp.in")
udp_out_logger = logging.getLogger("supriya.udp.out")

BUNDLE_PREFIX = b"#bundle\x00"
IMMEDIATELY = struct.pack(">Q", 1)
NTP_TIMESTAMP_TO_SECONDS = 1.0 / 2.0**32.0
SECONDS_TO_NTP_TIMESTAMP = 2.0**32.0
SYSTEM_EPOCH = datetime.date(*time.gmtime(0)[0:3])
NTP_EPOCH = datetime.date(1900, 1, 1)
NTP_DELTA = (SYSTEM_EPOCH - NTP_EPOCH).days * 24 * 3600


class OscMessage:
    """
    An OSC message.

    .. container:: example

        ::

            >>> from supriya.osc import OscMessage
            >>> osc_message = OscMessage("/g_new", 0, 0)
            >>> osc_message
            OscMessage('/g_new', 0, 0)

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/g_new', 0, 0)

        ::

            >>> print(osc_message)
            size 20
               0   2f 67 5f 6e  65 77 00 00  2c 69 69 00  00 00 00 00   |/g_new..,ii.....|
              16   00 00 00 00                                          |....|

    .. container:: example

        ::

            >>> osc_message = OscMessage("/foo", True, [None, [3.25]], OscMessage("/bar"))
            >>> osc_message
            OscMessage('/foo', True, [None, [3.25]], OscMessage('/bar'))

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/foo', True, [None, [3.25]], OscMessage('/bar'))

        ::

            >>> print(osc_message)
            size 40
               0   2f 66 6f 6f  00 00 00 00  2c 54 5b 4e  5b 66 5d 5d   |/foo....,T[N[f]]|
              16   62 00 00 00  40 50 00 00  00 00 00 0c  2f 62 61 72   |b...@P....../bar|
              32   00 00 00 00  2c 00 00 00                             |....,...|

    .. container:: example

        ::

            >>> osc_message = supriya.osc.OscMessage(
            ...     "/foo",
            ...     1,
            ...     2.5,
            ...     supriya.osc.OscBundle(
            ...         contents=(
            ...             supriya.osc.OscMessage("/bar", "baz", 3.0),
            ...             supriya.osc.OscMessage("/ffff", False, True, None),
            ...         )
            ...     ),
            ...     ["a", "b", ["c", "d"]],
            ... )
            >>> osc_message
            OscMessage('/foo', 1, 2.5, OscBundle(
                contents=(
                    OscMessage('/bar', 'baz', 3.0),
                    OscMessage('/ffff', False, True, None),
                ),
            ), ['a', 'b', ['c', 'd']])

        ::

            >>> datagram = osc_message.to_datagram()
            >>> OscMessage.from_datagram(datagram)
            OscMessage('/foo', 1, 2.5, OscBundle(
                contents=(
                    OscMessage('/bar', 'baz', 3.0),
                    OscMessage('/ffff', False, True, None),
                ),
            ), ['a', 'b', ['c', 'd']])

        ::

            >>> print(osc_message)
            size 112
               0   2f 66 6f 6f  00 00 00 00  2c 69 66 62  5b 73 73 5b   |/foo....,ifb[ss[|
              16   73 73 5d 5d  00 00 00 00  00 00 00 01  40 20 00 00   |ss]]........@ ..|
              32   00 00 00 3c  23 62 75 6e  64 6c 65 00  00 00 00 00   |...<#bundle.....|
              48   00 00 00 01  00 00 00 14  2f 62 61 72  00 00 00 00   |......../bar....|
              64   2c 73 66 00  62 61 7a 00  40 40 00 00  00 00 00 10   |,sf.baz.@@......|
              80   2f 66 66 66  66 00 00 00  2c 46 54 4e  00 00 00 00   |/ffff...,FTN....|
              96   61 00 00 00  62 00 00 00  63 00 00 00  64 00 00 00   |a...b...c...d...|
    """

    ### INITIALIZER ###

    def __init__(self, address, *contents) -> None:
        if isinstance(address, enum.Enum):
            address = address.value
        if not isinstance(address, (str, int)):
            raise ValueError(f"address must be int or str, got {address}")
        self.address = address
        self.contents = tuple(contents)

    ### SPECIAL METHODS ###

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return False
        if self.address != other.address:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        return "{}({})".format(
            type(self).__name__,
            ", ".join(repr(_) for _ in [self.address, *self.contents]),
        )

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    ### PRIVATE METHODS ###

    @staticmethod
    def _decode_blob(data):
        actual_length, remainder = struct.unpack(">I", data[:4])[0], data[4:]
        padded_length = actual_length
        if actual_length % 4 != 0:
            padded_length = (actual_length // 4 + 1) * 4
        return remainder[:padded_length][:actual_length], remainder[padded_length:]

    @staticmethod
    def _decode_string(data):
        actual_length = data.index(b"\x00")
        padded_length = (actual_length // 4 + 1) * 4
        return str(data[:actual_length], "ascii"), data[padded_length:]

    @staticmethod
    def _encode_string(value):
        result = bytes(value + "\x00", "ascii")
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @staticmethod
    def _encode_blob(value):
        result = bytes(struct.pack(">I", len(value)) + value)
        if len(result) % 4 != 0:
            width = (len(result) // 4 + 1) * 4
            result = result.ljust(width, b"\x00")
        return result

    @classmethod
    def _encode_value(cls, value):
        if hasattr(value, "to_datagram"):
            value = bytearray(value.to_datagram())
        elif isinstance(value, enum.Enum):
            value = value.value
        type_tags, encoded_value = "", b""
        if isinstance(value, (bytearray, bytes)):
            type_tags += "b"
            encoded_value = cls._encode_blob(value)
        elif isinstance(value, str):
            type_tags += "s"
            encoded_value = cls._encode_string(value)
        elif isinstance(value, bool):
            type_tags += "T" if value else "F"
        elif isinstance(value, float):
            type_tags += "f"
            encoded_value += struct.pack(">f", value)
        elif isinstance(value, int):
            type_tags += "i"
            encoded_value += struct.pack(">i", value)
        elif value is None:
            type_tags += "N"
        elif isinstance(value, SequenceABC):
            type_tags += "["
            for sub_value in value:
                sub_type_tags, sub_encoded_value = cls._encode_value(sub_value)
                type_tags += sub_type_tags
                encoded_value += sub_encoded_value
            type_tags += "]"
        else:
            message = "Cannot encode {!r}".format(value)
            raise TypeError(message)
        return type_tags, encoded_value

    ### PUBLIC METHODS ###

    def to_datagram(self) -> bytes:
        # address can be a string or (in SuperCollider) an int
        if isinstance(self.address, str):
            encoded_address = self._encode_string(self.address)
        else:
            encoded_address = struct.pack(">i", self.address)
        encoded_type_tags = ","
        encoded_contents = b""
        for value in self.contents or ():
            type_tags, encoded_value = self._encode_value(value)
            encoded_type_tags += type_tags
            encoded_contents += encoded_value
        return (
            encoded_address + self._encode_string(encoded_type_tags) + encoded_contents
        )

    @classmethod
    def from_datagram(cls, datagram):
        remainder = datagram
        address, remainder = cls._decode_string(remainder)
        type_tags, remainder = cls._decode_string(remainder)
        contents = []
        array_stack = [contents]
        for type_tag in type_tags[1:]:
            if type_tag == "i":
                value, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "f":
                value, remainder = struct.unpack(">f", remainder[:4])[0], remainder[4:]
                array_stack[-1].append(value)
            elif type_tag == "d":
                value, remainder = struct.unpack(">d", remainder[:8])[0], remainder[8:]
                array_stack[-1].append(value)
            elif type_tag == "s":
                value, remainder = cls._decode_string(remainder)
                array_stack[-1].append(value)
            elif type_tag == "b":
                value, remainder = cls._decode_blob(remainder)
                for class_ in (OscBundle, OscMessage):
                    try:
                        value = class_.from_datagram(value)
                        break
                    except Exception:
                        pass
                array_stack[-1].append(value)
            elif type_tag == "T":
                array_stack[-1].append(True)
            elif type_tag == "F":
                array_stack[-1].append(False)
            elif type_tag == "N":
                array_stack[-1].append(None)
            elif type_tag == "[":
                array = []
                array_stack[-1].append(array)
                array_stack.append(array)
            elif type_tag == "]":
                array_stack.pop()
            else:
                raise RuntimeError(f"Unable to parse type {type_tag!r}")
        return cls(address, *contents)

    def to_list(self):
        result = [self.address]
        for x in self.contents:
            if hasattr(x, "to_list"):
                result.append(x.to_list())
            else:
                result.append(x)
        return result


class OscBundle:
    """
    An OSC bundle.

    ::

        >>> import supriya.osc
        >>> message_one = supriya.osc.OscMessage("/one", 1)
        >>> message_two = supriya.osc.OscMessage("/two", 2)
        >>> message_three = supriya.osc.OscMessage("/three", 3)

    ::

        >>> inner_bundle = supriya.osc.OscBundle(
        ...     timestamp=1401557034.5,
        ...     contents=(message_one, message_two),
        ... )
        >>> inner_bundle
        OscBundle(
            contents=(
                OscMessage('/one', 1),
                OscMessage('/two', 2),
            ),
            timestamp=1401557034.5,
        )

    ::

        >>> print(inner_bundle)
        size 56
           0   23 62 75 6e  64 6c 65 00  d7 34 8e aa  80 00 00 00   |#bundle..4......|
          16   00 00 00 10  2f 6f 6e 65  00 00 00 00  2c 69 00 00   |..../one....,i..|
          32   00 00 00 01  00 00 00 10  2f 74 77 6f  00 00 00 00   |......../two....|
          48   2c 69 00 00  00 00 00 02                             |,i......|

    ::

        >>> outer_bundle = supriya.osc.OscBundle(
        ...     contents=(inner_bundle, message_three),
        ... )
        >>> outer_bundle
        OscBundle(
            contents=(
                OscBundle(
                    contents=(
                        OscMessage('/one', 1),
                        OscMessage('/two', 2),
                    ),
                    timestamp=1401557034.5,
                ),
                OscMessage('/three', 3),
            ),
        )

    ::

        >>> print(outer_bundle)
        size 96
           0   23 62 75 6e  64 6c 65 00  00 00 00 00  00 00 00 01   |#bundle.........|
          16   00 00 00 38  23 62 75 6e  64 6c 65 00  d7 34 8e aa   |...8#bundle..4..|
          32   80 00 00 00  00 00 00 10  2f 6f 6e 65  00 00 00 00   |......../one....|
          48   2c 69 00 00  00 00 00 01  00 00 00 10  2f 74 77 6f   |,i........../two|
          64   00 00 00 00  2c 69 00 00  00 00 00 02  00 00 00 10   |....,i..........|
          80   2f 74 68 72  65 65 00 00  2c 69 00 00  00 00 00 03   |/three..,i......|

    ::

        >>> datagram = outer_bundle.to_datagram()

    ::

        >>> decoded_bundle = supriya.osc.OscBundle.from_datagram(datagram)
        >>> decoded_bundle
        OscBundle(
            contents=(
                OscBundle(
                    contents=(
                        OscMessage('/one', 1),
                        OscMessage('/two', 2),
                    ),
                    timestamp=1401557034.5,
                ),
                OscMessage('/three', 3),
            ),
        )

    ::

        >>> decoded_bundle == outer_bundle
        True
    """

    ### INITIALIZER ###

    def __init__(self, timestamp=None, contents=None) -> None:
        prototype = (OscMessage, type(self))
        self.timestamp = timestamp
        contents = contents or ()
        for x in contents or ():
            if not isinstance(x, prototype):
                raise ValueError(contents)
        self.contents = tuple(contents)

    ### SPECIAL METHODS ###

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return False
        if self.timestamp != other.timestamp:
            return False
        if self.contents != other.contents:
            return False
        return True

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return format_datagram(bytearray(self.to_datagram()))

    ### PRIVATE METHODS ###

    @staticmethod
    def _decode_date(data):
        data, remainder = data[:8], data[8:]
        if data == IMMEDIATELY:
            return None, remainder
        date = (struct.unpack(">Q", data)[0] / SECONDS_TO_NTP_TIMESTAMP) - NTP_DELTA
        return date, remainder

    @staticmethod
    def _encode_date(seconds, realtime=True):
        if seconds is None:
            return IMMEDIATELY
        if realtime:
            seconds = seconds + NTP_DELTA
        if seconds >= 4294967296:  # 2**32
            seconds = seconds % 4294967296
        return struct.pack(">Q", int(seconds * SECONDS_TO_NTP_TIMESTAMP))

    ### PUBLIC METHODS ###

    @classmethod
    def from_datagram(cls, datagram):
        if not datagram.startswith(BUNDLE_PREFIX):
            raise ValueError("datagram is not a bundle")
        remainder = datagram[8:]
        timestamp, remainder = cls._decode_date(remainder)
        contents = []
        while len(remainder):
            length, remainder = struct.unpack(">i", remainder[:4])[0], remainder[4:]
            if remainder.startswith(BUNDLE_PREFIX):
                item = cls.from_datagram(remainder[:length])
            else:
                item = OscMessage.from_datagram(remainder[:length])
            contents.append(item)
            remainder = remainder[length:]
        osc_bundle = cls(timestamp=timestamp, contents=tuple(contents))
        return osc_bundle

    @classmethod
    def partition(cls, messages, timestamp=None):
        bundles = []
        contents = []
        message = collections.deque(messages)
        remaining = maximum = 8192 - len(BUNDLE_PREFIX) - 4
        while messages:
            message = messages.popleft()
            datagram = message.to_datagram()
            remaining -= len(datagram) + 4
            if remaining > 0:
                contents.append(message)
            else:
                bundles.append(cls(timestamp=timestamp, contents=contents))
                contents = [message]
                remaining = maximum
        if contents:
            bundles.append(cls(timestamp=timestamp, contents=contents))
        return bundles

    def to_datagram(self, realtime=True) -> bytes:
        datagram = BUNDLE_PREFIX
        datagram += self._encode_date(self.timestamp, realtime=realtime)
        for content in self.contents:
            content_datagram = content.to_datagram()
            datagram += struct.pack(">i", len(content_datagram))
            datagram += content_datagram
        return datagram

    def to_list(self):
        result = [self.timestamp]
        result.append([x.to_list() for x in self.contents])
        return result


class OscProtocolOffline(Exception):
    pass


class OscProtocolAlreadyConnected(Exception):
    pass


class OscCallback(NamedTuple):
    pattern: Tuple[Union[str, int, float], ...]
    procedure: Callable
    failure_pattern: Optional[Tuple[Union[str, int, float], ...]] = None
    once: bool = False


@dataclasses.dataclass
class HealthCheck:
    request_pattern: List[str]
    response_pattern: List[str]
    callback: Callable
    active: bool = True
    timeout: float = 1.0
    backoff_factor: float = 1.5
    max_attempts: int = 5


class OscProtocol(metaclass=abc.ABCMeta):
    ### INITIALIZER ###

    def __init__(self) -> None:
        self.callbacks: Dict[Any, Any] = {}
        self.captures: Set[Capture] = set()
        self.healthcheck = None
        self.healthcheck_osc_callback = None
        self.attempts = 0
        self.ip_address = None
        self.is_running: bool = False
        self.port = None

    ### PRIVATE METHODS ###

    def _add_callback(self, callback: OscCallback):
        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            callback_map = self.callbacks
            for item in pattern:
                callbacks, callback_map = callback_map.setdefault(item, ([], {}))
            callbacks.append(callback)

    def _disconnect(self) -> None:
        raise NotImplementedError

    def _match_callbacks(self, message):
        items = (message.address,) + message.contents
        matching_callbacks = []
        callback_map = self.callbacks
        for item in items:
            if item not in callback_map:
                break
            callbacks, callback_map = callback_map[item]
            matching_callbacks.extend(callbacks)
        for callback in matching_callbacks:
            if callback.once:
                self.unregister(callback)
        return matching_callbacks

    def _remove_callback(self, callback: OscCallback):
        def delete(pattern, original_callback_map):
            key = pattern.pop(0)
            if key not in original_callback_map:
                return
            callbacks, callback_map = original_callback_map[key]
            if pattern:
                delete(pattern, callback_map)
            if callback in callbacks:
                callbacks.remove(callback)
            if not callbacks and not callback_map:
                original_callback_map.pop(key)

        patterns = [callback.pattern]
        if callback.failure_pattern:
            patterns.append(callback.failure_pattern)
        for pattern in patterns:
            delete(list(pattern), self.callbacks)

    def _pass_healthcheck(self, message):
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] healthcheck: passed")
        self.attempts = 0

    def _setup(self, ip_address, port, healthcheck):
        self.ip_address = ip_address
        self.port = port
        self.healthcheck = healthcheck
        if self.healthcheck:
            self.healthcheck_osc_callback = self.register(
                pattern=self.healthcheck.response_pattern,
                procedure=self._pass_healthcheck,
            )

    def _teardown(self):
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] Tearing down...")
        self.is_running = False
        if self.healthcheck is not None:
            self.unregister(self.healthcheck_osc_callback)

    def _validate_callback(
        self, pattern, procedure, *, failure_pattern=None, once=False
    ):
        if isinstance(pattern, (str, int, float)):
            pattern = [pattern]
        if isinstance(failure_pattern, (str, int, float)):
            failure_pattern = [failure_pattern]
        if not callable(procedure):
            raise ValueError(procedure)
        return OscCallback(
            pattern=tuple(pattern),
            failure_pattern=failure_pattern,
            procedure=procedure,
            once=bool(once),
        )

    def _validate_receive(self, datagram):
        udp_in_logger.debug(f"[{self.ip_address}:{self.port}] {datagram}")
        try:
            message = OscMessage.from_datagram(datagram)
        except Exception:
            raise
        osc_in_logger.debug(f"[{self.ip_address}:{self.port}] {message!r}")
        for capture in self.captures:
            capture.messages.append(
                CaptureEntry(timestamp=time.time(), label="R", message=message)
            )
        for callback in self._match_callbacks(message):
            yield callback.procedure, message

    def _validate_send(self, message):
        if not self.is_running:
            raise OscProtocolOffline
        if not isinstance(message, (str, SequenceABC, OscBundle, OscMessage)):
            raise ValueError(message)
        if isinstance(message, str):
            message = OscMessage(message)
        elif isinstance(message, SequenceABC):
            message = OscMessage(*message)
        osc_out_logger.debug(f"[{self.ip_address}:{self.port}] {message!r}")
        for capture in self.captures:
            capture.messages.append(
                CaptureEntry(timestamp=time.time(), label="S", message=message)
            )
        datagram = message.to_datagram()
        udp_out_logger.debug(f"[{self.ip_address}:{self.port}] {datagram}")
        return datagram

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def activate_healthcheck(self) -> None:
        raise NotImplementedError

    def capture(self) -> "Capture":
        return Capture(self)

    def disconnect(self) -> None:
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] disconnecting")
        self._disconnect()
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] ...disconnected")

    @abc.abstractmethod
    def register(
        self,
        pattern: Sequence[Union[str, float]],
        procedure: Callable[[OscMessage], None],
        *,
        failure_pattern: Optional[Sequence[Union[str, float]]] = None,
        once: bool = False,
    ) -> OscCallback:
        raise NotImplementedError

    @abc.abstractmethod
    def send(self, message) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def unregister(self, callback: OscCallback):
        raise NotImplementedError


class AsyncOscProtocol(asyncio.DatagramProtocol, OscProtocol):
    ### INITIALIZER ###

    def __init__(self) -> None:
        asyncio.DatagramProtocol.__init__(self)
        OscProtocol.__init__(self)
        self.background_tasks: Set[asyncio.Task] = set()
        self.healthcheck_task: Optional[asyncio.Task] = None

    ### PRIVATE METHODS ###

    def _disconnect(self) -> None:
        if not self.is_running:
            osc_protocol_logger.info(
                f"{self.ip_address}:{self.port} already disconnected!"
            )
            return
        self._teardown()
        self.transport.close()
        if self.healthcheck_task:
            self.healthcheck_task.cancel()

    async def _run_healthcheck(self):
        while self.is_running:
            if self.attempts >= self.healthcheck.max_attempts:
                osc_protocol_logger.info(
                    f"[{self.ip_address}:{self.port}] health check: failure limit exceeded"
                )
                self.exit_future.set_result(True)
                self._teardown()
                self.transport.close()
                obj_ = self.healthcheck.callback()
                if asyncio.iscoroutine(obj_):
                    asyncio.get_running_loop().create_task(obj_)
                return
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}] healthcheck: checking..."
            )
            self.send(OscMessage(*self.healthcheck.request_pattern))
            sleep_time = self.healthcheck.timeout * pow(
                self.healthcheck.backoff_factor, self.attempts
            )
            self.attempts += 1
            await asyncio.sleep(sleep_time)

    ### PUBLIC METHODS ###

    def activate_healthcheck(self) -> None:
        if not self.healthcheck:
            return
        elif self.healthcheck.active:
            return
        self.healthcheck_task = asyncio.get_running_loop().create_task(
            self._run_healthcheck()
        )

    async def connect(
        self, ip_address: str, port: int, *, healthcheck: Optional[HealthCheck] = None
    ):
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] connecting...")
        if self.is_running:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}] already connected!"
            )
            raise OscProtocolAlreadyConnected
        self._setup(ip_address, port, healthcheck)
        loop = asyncio.get_running_loop()
        self.exit_future = loop.create_future()
        _, protocol = await loop.create_datagram_endpoint(
            lambda: self, remote_addr=(ip_address, port)
        )
        if self.healthcheck and self.healthcheck.active:
            self.healthcheck_task = asyncio.get_running_loop().create_task(
                self._run_healthcheck()
            )
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] ...connected")

    def connection_made(self, transport):
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] connection made")
        self.transport = transport
        self.is_running = True

    def connection_lost(self, exc):
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] connection lost")
        self.exit_future.set_result(True)

    def datagram_received(self, data, addr):
        loop = asyncio.get_running_loop()
        for callback, message in self._validate_receive(data):
            if inspect.iscoroutinefunction(callback):
                task = loop.create_task(callback(message))
                self.background_tasks.add(task)
                task.add_done_callback(self.background_tasks.discard)
            else:
                callback(message)

    def error_received(self, exc):
        osc_out_logger.warning(f"[{self.ip_address}:{self.port}] errored: {exc}")

    def register(
        self,
        pattern: Sequence[Union[str, float]],
        procedure: Callable[[OscMessage], None],
        *,
        failure_pattern: Optional[Sequence[Union[str, float]]] = None,
        once: bool = False,
    ) -> OscCallback:
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}] registering pattern: {pattern!r}"
        )
        callback = self._validate_callback(
            pattern, procedure, failure_pattern=failure_pattern, once=once
        )
        self._add_callback(callback)
        return callback

    def send(self, message):
        osc_protocol_logger.debug(
            f"[{self.ip_address}:{self.port}] sending: {message!r}"
        )
        datagram = self._validate_send(message)
        return self.transport.sendto(datagram)

    def unregister(self, callback: OscCallback):
        self._remove_callback(callback)


class ThreadedOscServer(socketserver.UDPServer):
    osc_protocol: "ThreadedOscProtocol"

    def verify_request(self, request, client_address):
        self.osc_protocol._process_command_queue()
        return True

    def service_actions(self):
        if self.osc_protocol.healthcheck.active:
            self.osc_protocol._run_healthcheck()


class ThreadedOscHandler(socketserver.BaseRequestHandler):
    def handle(self):
        data = self.request[0]
        for procedure, message in self.server.osc_protocol._validate_receive(data):
            procedure(message)


class ThreadedOscProtocol(OscProtocol):
    ### INITIALIZER ###

    def __init__(self):
        OscProtocol.__init__(self)
        self.command_queue = queue.Queue()
        self.lock = threading.RLock()
        self.osc_server = None
        self.osc_server_thread = None

    ### PRIVATE METHODS ###

    def _disconnect(self) -> None:
        with self.lock:
            if not self.is_running:
                osc_protocol_logger.info(
                    f"{self.ip_address}:{self.port} already disconnected!"
                )
                return
            self._teardown()
            if not self.osc_server._BaseServer__shutdown_request:
                self.osc_server.shutdown()
            self.osc_server = None
            self.osc_server_thread = None

    def _process_command_queue(self):
        while self.command_queue.qsize():
            try:
                action, callback = self.command_queue.get()
            except queue.Empty:
                continue
            if action == "add":
                self._add_callback(callback)
            elif action == "remove":
                self._remove_callback(callback)

    def _run_healthcheck(self):
        if self.healthcheck is None:
            return
        now = time.time()
        if now < self.healthcheck_deadline:
            return
        if self.attempts > 0:
            remaining = self.healthcheck.max_attempts - self.attempts
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}] healthcheck failed, {remaining} attempts remaining"
            )
        new_timeout = self.healthcheck.timeout * pow(
            self.healthcheck.backoff_factor, self.attempts
        )
        self.healthcheck_deadline = now + new_timeout
        self.attempts += 1
        if self.attempts <= self.healthcheck.max_attempts:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}] healthcheck: checking..."
            )
            self.send(OscMessage(*self.healthcheck.request_pattern))
            return
        osc_protocol_logger.info(
            f"[{self.ip_address}:{self.port}] healthcheck: failure limit exceeded"
        )
        self.osc_server._BaseServer__shutdown_request = True
        self.disconnect()
        self.healthcheck.callback()

    def _server_factory(self, ip_address, port):
        server = ThreadedOscServer(
            (self.ip_address, self.port), ThreadedOscHandler, bind_and_activate=False
        )
        server.osc_protocol = self
        return server

    ### PUBLIC METHODS ###

    def activate_healthcheck(self) -> None:
        if not self.healthcheck:
            return
        elif self.healthcheck.active:
            return
        self.healthcheck.active = True

    def connect(
        self, ip_address: str, port: int, *, healthcheck: Optional[HealthCheck] = None
    ):
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] connecting...")
        if self.is_running:
            osc_protocol_logger.info(
                f"[{self.ip_address}:{self.port}] already connected!"
            )
            raise OscProtocolAlreadyConnected
        self._setup(ip_address, port, healthcheck)
        self.healthcheck_deadline = time.time()
        self.osc_server = self._server_factory(ip_address, port)
        self.osc_server_thread = threading.Thread(target=self.osc_server.serve_forever)
        self.osc_server_thread.daemon = True
        self.osc_server_thread.start()
        self.is_running = True
        osc_protocol_logger.info(f"[{self.ip_address}:{self.port}] ...connected")

    def register(
        self,
        pattern: Sequence[Union[str, float]],
        procedure: Callable[[OscMessage], None],
        *,
        failure_pattern: Optional[Sequence[Union[str, float]]] = None,
        once: bool = False,
    ) -> OscCallback:
        """
        Register a callback.
        """
        callback = self._validate_callback(
            pattern, procedure, failure_pattern=failure_pattern, once=once
        )
        # Command queue prevents lock contention.
        self.command_queue.put(("add", callback))
        return callback

    def send(self, message) -> None:
        datagram = self._validate_send(message)
        try:
            self.osc_server.socket.sendto(datagram, (self.ip_address, self.port))
        except OSError:
            # print(message)
            raise

    def unregister(self, callback: OscCallback) -> None:
        """
        Unregister a callback.
        """
        # Command queue prevents lock contention.
        self.command_queue.put(("remove", callback))


class CaptureEntry(NamedTuple):
    timestamp: float
    label: str
    message: Union[OscMessage, OscBundle]


class Capture:
    ### INITIALIZER ###

    def __init__(self, osc_protocol):
        self.osc_protocol = osc_protocol
        self.messages = []

    ### SPECIAL METHODS ###

    def __enter__(self):
        self.osc_protocol.captures.add(self)
        self.messages[:] = []
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.osc_protocol.captures.remove(self)

    def __iter__(self):
        return iter(self.messages)

    def __len__(self):
        return len(self.messages)

    ### PUBLIC METHODS ###

    def filtered(
        self, sent=True, received=True, status=True
    ) -> List[Union[OscBundle, OscMessage]]:
        messages = []
        for _, label, message in self.messages:
            if label == "R" and not received:
                continue
            if label == "S" and not sent:
                continue
            if (
                isinstance(message, OscMessage)
                and message.address in ("/status", "/status.reply")
                and not status
            ):
                continue
            messages.append(message)
        return messages

    ### PUBLIC PROPERTIES ###

    @property
    def received_messages(self):
        return [
            (timestamp, osc_message)
            for timestamp, label, osc_message in self.messages
            if label == "R"
        ]

    @property
    def sent_messages(self):
        return [
            (timestamp, osc_message)
            for timestamp, label, osc_message in self.messages
            if label == "S"
        ]


def find_free_port():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_DGRAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def format_datagram(datagram):
    result = []
    result.append("size {}".format(len(datagram)))
    index = 0
    while index < len(datagram):
        chunk = datagram[index : index + 16]
        line = "{: >4}   ".format(index)
        hex_blocks = []
        ascii_block = ""
        for chunk in group_iterable_by_count(chunk, 4):
            hex_block = []
            for byte in chunk:
                char = int(byte)
                if 31 < char < 127:
                    char = chr(char)
                else:
                    char = "."
                ascii_block += char
                hexed = hex(byte)[2:].zfill(2)
                hex_block.append(hexed)
            hex_block = " ".join(hex_block)
            hex_blocks.append(hex_block)
        hex_blocks = "  ".join(hex_blocks)
        ascii_block = "|{}|".format(ascii_block)
        hex_blocks = "{: <53}".format(hex_blocks)
        line += hex_blocks
        line += ascii_block
        result.append(line)
        index += 16
    result = "\n".join(result)
    return result


__all__ = [
    "AsyncOscProtocol",
    "Capture",
    "CaptureEntry",
    "HealthCheck",
    "OscBundle",
    "OscCallback",
    "OscMessage",
    "OscProtocol",
    "ThreadedOscProtocol",
    "find_free_port",
]



"""
Classes for modeling requests to :term:`scsynth`.
"""

import asyncio
import dataclasses
import logging
from abc import ABC, abstractmethod
from concurrent.futures import Future
from os import PathLike
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Sequence,
    SupportsFloat,
    SupportsInt,
    Tuple,
    Union,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from uqbar.objects import new

from ..enums import AddAction, HeaderFormat, RequestName, SampleFormat
from ..osc import OscBundle, OscMessage
from ..synthdefs import SynthDef, SynthDefCompiler
from ..typing import AddActionLike, HeaderFormatLike, SampleFormatLike, SupportsOsc
from .responses import Response

if TYPE_CHECKING:
    from .contexts.core import Context
    from .contexts.realtime import AsyncServer, Server


logger = logging.getLogger(__name__)


class Requestable(ABC):
    """
    Abstract base for request-like classes.
    """

    ### PRIVATE METHODS ###

    @abstractmethod
    def _get_response_patterns_and_requestable(
        self, context: "Context"
    ) -> Tuple[
        Optional[Sequence[Union[float, str]]],
        Optional[Sequence[Union[float, str]]],
        "Requestable",
    ]:
        raise NotImplementedError

    ### PUBLIC METHODS ###

    def communicate(self, server: "Server", timeout: float = 1.0) -> Optional[Response]:
        (
            success_pattern,
            failure_pattern,
            requestable,
        ) = self._get_response_patterns_and_requestable(server)
        if not success_pattern:
            server.send(self)
            return None
        future: Future[Response] = Future()
        server._osc_protocol.register(
            pattern=success_pattern,
            failure_pattern=failure_pattern,
            procedure=lambda message: future.set_result(Response.from_osc(message)),
            once=True,
        )
        server.send(requestable)
        return future.result(timeout=timeout)

    async def communicate_async(
        self, server: "AsyncServer", timeout: float = 1.0
    ) -> Optional[Response]:
        (
            success_pattern,
            failure_pattern,
            requestable,
        ) = self._get_response_patterns_and_requestable(server)
        if not success_pattern:
            server.send(self)
            return None
        future: asyncio.Future[Response] = asyncio.get_running_loop().create_future()
        osc_callback = server._osc_protocol.register(
            pattern=success_pattern,
            failure_pattern=failure_pattern,
            procedure=lambda message: future.set_result(Response.from_osc(message)),
            once=True,
        )
        server.send(requestable)
        try:
            await asyncio.wait_for(future, timeout=timeout)
        except asyncio.TimeoutError:
            # No response received, so make sure to cleanup
            server._osc_protocol.unregister(osc_callback)
            raise
        return future.result()

    @abstractmethod
    def to_osc(self) -> Union[OscBundle, OscMessage]:
        raise NotImplementedError


@dataclasses.dataclass
class Request(Requestable):
    """
    Abstract base for request classes.
    """

    ### PRIVATE METHODS ###

    def _get_response_patterns(
        self,
    ) -> Tuple[
        Optional[Sequence[Union[float, str]]], Optional[Sequence[Union[float, str]]]
    ]:
        return None, None

    def _get_response_patterns_and_requestable(
        self, context: "Context"
    ) -> Tuple[
        Optional[Sequence[Union[float, str]]],
        Optional[Sequence[Union[float, str]]],
        "Requestable",
    ]:
        success_pattern, failure_pattern = self._get_response_patterns()
        return success_pattern, failure_pattern, self

    ### PUBLIC METHODS ###

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        return requests


@dataclasses.dataclass
class RequestBundle(Requestable):
    """
    A bundle of requestables to be executed at a timestamp.
    """

    contents: Sequence[Requestable]
    timestamp: Optional[float] = None

    ### PRIVATE METHODS ###

    def _get_response_patterns_and_requestable(
        self, context: "Context"
    ) -> Tuple[
        Optional[Sequence[Union[float, str]]],
        Optional[Sequence[Union[float, str]]],
        "Requestable",
    ]:
        sync_id = context._get_next_sync_id()
        request_bundle: "RequestBundle" = new(
            self, contents=list(self.contents) + [Sync(sync_id=sync_id)]
        )
        response_pattern: List[Union[float, str]] = ["/synced", sync_id]
        return response_pattern, None, request_bundle

    ### PUBLIC METHODS ###

    def to_osc(self) -> OscBundle:
        return OscBundle(
            contents=[x.to_osc() for x in self.contents], timestamp=self.timestamp
        )


@dataclasses.dataclass
class AllocateBuffer(Request):
    """
    A ``/b_alloc`` request.

    ::

        >>> from supriya.contexts.requests import AllocateBuffer, NormalizeBuffer
        >>> request = AllocateBuffer(
        ...     buffer_id=1,
        ...     frame_count=512,
        ...     channel_count=1,
        ...     on_completion=NormalizeBuffer(buffer_id=1),
        ... )
        >>> request.to_osc()
        OscMessage('/b_alloc', 1, 512, 1, OscMessage('/b_gen', 1, 'normalize', 1.0))
    """

    buffer_id: SupportsInt
    frame_count: int
    channel_count: int = 1
    on_completion: Optional[SupportsOsc] = None

    def _get_response_patterns(self):
        return ["/done", "/b_alloc", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int]] = [
            int(self.buffer_id),
            int(self.frame_count),
            int(self.channel_count),
        ]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_ALLOCATE, *contents)


@dataclasses.dataclass
class AllocateReadBuffer(Request):
    """
    A ``/b_allocRead`` request.

    ::

        >>> from supriya.contexts.requests import AllocateReadBuffer, NormalizeBuffer
        >>> request = AllocateReadBuffer(
        ...     buffer_id=1,
        ...     path="path/to/audio.wav",
        ...     starting_frame=32,
        ...     frame_count=512,
        ...     on_completion=NormalizeBuffer(buffer_id=1),
        ... )
        >>> request.to_osc()
        OscMessage('/b_allocRead', 1, 'path/to/audio.wav', 32, 512, OscMessage('/b_gen', 1, 'normalize', 1.0))
    """

    buffer_id: SupportsInt
    path: PathLike
    starting_frame: int = 0
    frame_count: int = 0
    on_completion: Optional[SupportsOsc] = None

    def _get_response_patterns(self):
        return ["/done", "/b_allocRead", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int, str]] = [
            int(self.buffer_id),
            str(self.path),
            int(self.starting_frame),
            int(self.frame_count),
        ]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_ALLOCATE_READ, *contents)


@dataclasses.dataclass
class AllocateReadBufferChannel(Request):
    """
    A ``/b_allocReadChannel`` request.

    ::

        >>> from supriya.contexts.requests import AllocateReadBufferChannel, NormalizeBuffer
        >>> request = AllocateReadBufferChannel(
        ...     buffer_id=1,
        ...     path="path/to/audio.wav",
        ...     channel_indices=[1, 2],
        ...     starting_frame=32,
        ...     frame_count=512,
        ...     on_completion=NormalizeBuffer(buffer_id=1),
        ... )
        >>> request.to_osc()
        OscMessage('/b_allocReadChannel', 1, 'path/to/audio.wav', 32, 512, 1, 2, OscMessage('/b_gen', 1, 'normalize', 1.0))
    """

    buffer_id: SupportsInt
    path: PathLike
    channel_indices: Sequence[int]
    starting_frame: int = 0
    frame_count: int = 0
    on_completion: Optional[SupportsOsc] = None

    def _get_response_patterns(self):
        return ["/done", "/b_allocReadChannel", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int, str]] = [
            int(self.buffer_id),
            str(self.path),
            int(self.starting_frame),
            int(self.frame_count),
            *(int(channel_index) for channel_index in self.channel_indices),
        ]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_ALLOCATE_READ_CHANNEL, *contents)


@dataclasses.dataclass
class AutoReassignSynthID(Request):
    """
    A ``/s_noid`` request.

    ::

        >>> from supriya.contexts.requests import AutoReassignSynthID
        >>> request = AutoReassignSynthID(synth_ids=[1, 2, 3])
        >>> request.to_osc()
        OscMessage('/s_noid', 1, 2, 3)
    """

    synth_ids: Sequence[SupportsInt]

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.SYNTH_NOID, *(int(synth_id) for synth_id in self.synth_ids)
        )


@dataclasses.dataclass
class ClearSchedule(Request):
    """
    A ``/clearSched`` request.

    ::

        >>> from supriya.contexts.requests import ClearSchedule
        >>> request = ClearSchedule()
        >>> request.to_osc()
        OscMessage('/clearSched')
    """

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.CLEAR_SCHEDULE)


@dataclasses.dataclass
class CloseBuffer(Request):
    """
    A ``/b_close`` request.

    ::

        >>> from supriya.contexts.requests import CloseBuffer, FreeNode
        >>> request = CloseBuffer(buffer_id=1, on_completion=FreeNode(node_ids=[1]))
        >>> request.to_osc()
        OscMessage('/b_close', 1, OscMessage('/n_free', 1))
    """

    buffer_id: SupportsInt
    on_completion: Optional[Requestable] = None

    def _get_response_patterns(self):
        return ["/done", "/b_close", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int]] = [int(self.buffer_id)]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_CLOSE, *contents)


@dataclasses.dataclass
class CopyBuffer(Request):
    """
    A ``/b_gen`` ``copy`` request.

    ::

        >>> from supriya.contexts.requests import CopyBuffer
        >>> request = CopyBuffer(
        ...     source_buffer_id=1,
        ...     target_buffer_id=2,
        ...     frame_count=512,
        ...     source_starting_frame=0,
        ...     target_starting_frame=128,
        ... )
        >>> request.to_osc()
        OscMessage('/b_gen', 2, 'copy', 128, 1, 0, 512)
    """

    source_buffer_id: SupportsInt
    target_buffer_id: SupportsInt
    frame_count: int = -1
    source_starting_frame: int = 0
    target_starting_frame: int = 0

    def _get_response_patterns(self):
        return ["/done", "/b_gen", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.BUFFER_GENERATE,
            int(self.target_buffer_id),
            "copy",
            int(self.target_starting_frame),
            int(self.source_buffer_id),
            int(self.source_starting_frame),
            int(self.frame_count),
        )


@dataclasses.dataclass
class DoNothing(Request):
    """
    A "nothing" request.

    ::

        >>> from supriya.contexts.requests import DoNothing
        >>> request = DoNothing()
        >>> request.to_osc()
        OscMessage(0)
    """

    def to_osc(self) -> OscMessage:
        return OscMessage(0)


@dataclasses.dataclass
class DumpOsc(Request):
    """
    A ``/dumpOSC`` request.

    ::

        >>> from supriya.contexts.requests import DumpOsc
        >>> request = DumpOsc(True)
        >>> request.to_osc()
        OscMessage('/dumpOSC', 1)
    """

    code: int

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.DUMP_OSC, int(self.code))


@dataclasses.dataclass
class DumpTree(Request):
    """
    A ``/g_dumpTree`` request.

    ::

        >>> from supriya.contexts.requests import DumpTree
        >>> request = DumpTree(items=[(0, True)])
        >>> request.to_osc()
        OscMessage('/g_dumpTree', 0, True)
    """

    items: Sequence[Tuple[SupportsInt, bool]]

    def to_osc(self) -> OscMessage:
        contents = []
        for node_id, flag in self.items:
            contents.extend([int(node_id), bool(flag)])
        return OscMessage(RequestName.GROUP_DUMP_TREE, *contents)


@dataclasses.dataclass
class FillBuffer(Request):
    """
    A ``/b_fill`` request.

    ::

        >>> from supriya.contexts.requests import FillBuffer
        >>> request = FillBuffer(
        ...     buffer_id=1,
        ...     items=[(32, 64, 0.5)],
        ... )
        >>> request.to_osc()
        OscMessage('/b_fill', 1, 32, 64, 0.5)
    """

    buffer_id: SupportsInt
    items: Sequence[Tuple[int, int, float]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items_by_buffer_id: Dict[int, List[Tuple[int, int, float]]] = {}
        for request in requests:
            if isinstance(request, cls):
                items_by_buffer_id.setdefault(int(request.buffer_id), []).extend(
                    request.items
                )
        return [
            cls(buffer_id=buffer_id, items=items)
            for buffer_id, items in sorted(items_by_buffer_id.items())
        ]

    def to_osc(self) -> OscMessage:
        contents: List[Union[float]] = [int(self.buffer_id)]
        for index, count, value in self.items:
            contents.extend([int(index), int(count), float(value)])
        return OscMessage(RequestName.BUFFER_FILL, *contents)


@dataclasses.dataclass
class FillControlBusRange(Request):
    """
    A ``/c_fill`` request.

    ::

        >>> from supriya.contexts.requests import FillControlBusRange
        >>> request = FillControlBusRange(
        ...     items=[(4, 12, 0.25), (32, 4, 440.0)],
        ... )
        >>> request.to_osc()
        OscMessage('/c_fill', 4, 12, 0.25, 32, 4, 440.0)
    """

    items: Sequence[Tuple[SupportsInt, int, float]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items: List[Tuple[SupportsInt, int, float]] = []
        for request in requests:
            if not isinstance(request, cls):
                continue
            items.extend(request.items)
        return [cls(items=items)]

    def to_osc(self) -> OscMessage:
        contents: List[float] = []
        for index, count, value in self.items:
            contents.extend([int(index), int(count), float(value)])
        return OscMessage(RequestName.CONTROL_BUS_FILL, *contents)


@dataclasses.dataclass
class FillNode(Request):
    """
    A ``/n_fill`` request.

    ::

        >>> from supriya.contexts.requests import FillNode
        >>> request = FillNode(
        ...     node_id=1000,
        ...     items=[("frequency", 7, 432.0)]
        ... )
        >>> request.to_osc()
        OscMessage('/n_fill', 1000, 'frequency', 7, 432.0)
    """

    node_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], int, float]]

    def to_osc(self) -> OscMessage:
        contents: List[Union[float, str]] = [int(self.node_id)]
        for control, count, value in self.items:
            contents.extend(
                [
                    control if isinstance(control, str) else int(control),
                    int(count),
                    float(value),
                ]
            )
        return OscMessage(RequestName.NODE_FILL, *contents)


@dataclasses.dataclass
class FreeAllSynthDefs(Request):
    """
    A ``/d_freeAll`` request.

    ::

        >>> from supriya.contexts.requests import FreeAllSynthDefs
        >>> request = FreeAllSynthDefs()
        >>> request.to_osc()
        OscMessage('/d_freeAll')
    """

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.SYNTHDEF_FREE_ALL)


@dataclasses.dataclass
class FreeBuffer(Request):
    """
    A ``/b_free`` request.

    ::

        >>> from supriya.contexts.requests import FreeBuffer, FreeNode
        >>> request = FreeBuffer(
        ...     buffer_id=1,
        ...     on_completion=FreeNode(node_ids=[1000]),
        ... )
        >>> request.to_osc()
        OscMessage('/b_free', 1, OscMessage('/n_free', 1000))
    """

    buffer_id: SupportsInt
    on_completion: Optional[Requestable] = None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int]] = [int(self.buffer_id)]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_FREE, *contents)


@dataclasses.dataclass
class FreeGroupChildren(Request):
    """
    A ``/g_freeAll`` request.

    ::

        >>> from supriya.contexts.requests import FreeGroupChildren
        >>> request = FreeGroupChildren(node_ids=[1])
        >>> request.to_osc()
        OscMessage('/g_freeAll', 1)
    """

    node_ids: Sequence[SupportsInt]

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.GROUP_FREE_ALL, *(int(node_id) for node_id in self.node_ids)
        )


@dataclasses.dataclass
class FreeGroupDeep(Request):
    """
    A ``/g_deepFree`` request.

    ::

        >>> from supriya.contexts.requests import FreeGroupDeep
        >>> request = FreeGroupDeep(node_ids=[1])
        >>> request.to_osc()
        OscMessage('/g_deepFree', 1)
    """

    node_ids: Sequence[SupportsInt]

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.GROUP_DEEP_FREE, *(int(node_id) for node_id in self.node_ids)
        )


@dataclasses.dataclass
class FreeNode(Request):
    """
    A ``/n_free`` request.

    ::

        >>> from supriya.contexts.requests import FreeNode
        >>> request = FreeNode(node_ids=[1000, 1001])
        >>> request.to_osc()
        OscMessage('/n_free', 1000, 1001)
    """

    node_ids: Sequence[SupportsInt]

    def _get_response_patterns(self):
        return ["/n_end", int(self.node_ids[-1])], None

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.NODE_FREE, *(int(node_id) for node_id in self.node_ids)
        )


@dataclasses.dataclass
class FreeSynthDef(Request):
    """
    A ``/d_free`` request.

    ::

        >>> from supriya.contexts.requests import FreeSynthDef
        >>> request = FreeSynthDef(synthdefs=["default"])
        >>> request.to_osc()
        OscMessage('/d_free', 'default')
    """

    synthdefs: Sequence[Union[SynthDef, str]]

    def to_osc(self) -> OscMessage:
        contents = []
        for x in self.synthdefs:
            if isinstance(x, SynthDef):
                contents.append(x.actual_name)
            else:
                contents.append(x)
        return OscMessage(RequestName.SYNTHDEF_FREE, *contents)


@dataclasses.dataclass
class GenerateBuffer(Request):
    """
    A ``/b_gen`` request, for use with ``sine1``, ``sine2``, ``sine3`` and ``cheby`` commands.

    ::

        >>> from supriya.contexts.requests import GenerateBuffer
        >>> request = GenerateBuffer(
        ...     buffer_id=1,
        ...     command_name="sine3",
        ...     amplitudes=[1, 2, 3],
        ...     frequencies=[4, 5, 6],
        ...     phases=[0.25, 0.5, 0.75],
        ...     should_normalize=True,
        ... )
        >>> request.to_osc()
        OscMessage('/b_gen', 1, 'sine3', 1, 4.0, 1.0, 0.25, 5.0, 2.0, 0.5, 6.0, 3.0, 0.75)
    """

    buffer_id: SupportsInt
    command_name: Literal["sine1", "sine2", "sine3", "cheby"]
    amplitudes: Sequence[float]
    frequencies: Optional[Sequence[float]]
    phases: Optional[Sequence[float]]
    as_wavetable: bool = False
    should_clear_first: bool = False
    should_normalize: bool = False

    def _get_response_patterns(self):
        return ["/done", "/b_gen", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[str, float]] = [
            int(self.buffer_id),
            self.command_name,
            (
                self.should_normalize
                | (self.as_wavetable * 2)
                | (self.should_clear_first * 4)
            ),
        ]
        sequences = []
        if self.frequencies:
            sequences.append(self.frequencies)
        sequences.append(self.amplitudes)
        if self.phases:
            sequences.append(self.phases)
        for values in zip(*sequences):
            for value in values:
                contents.append(float(value))
        return OscMessage(RequestName.BUFFER_GENERATE, *contents)


@dataclasses.dataclass
class GetBuffer(Request):
    """
    A ``/b_get`` request.

    ::

        >>> from supriya.contexts.requests import GetBuffer
        >>> request = GetBuffer(
        ...     buffer_id=1,
        ...     indices=[0, 4, 8],
        ... )
        >>> request.to_osc()
        OscMessage('/b_get', 1, 0, 4, 8)
    """

    buffer_id: SupportsInt
    indices: Sequence[int]

    def _get_response_patterns(self):
        return ["/b_set", int(self.buffer_id)], None

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.BUFFER_GET,
            int(self.buffer_id),
            *(int(index) for index in self.indices),
        )


@dataclasses.dataclass
class GetBufferRange(Request):
    """
    A ``/b_getn`` request.

    ::

        >>> from supriya.contexts.requests import GetBufferRange
        >>> request = GetBufferRange(
        ...     buffer_id=1,
        ...     items=[(0, 4), (32, 8)],
        ... )
        >>> request.to_osc()
        OscMessage('/b_getn', 1, 0, 4, 32, 8)
    """

    buffer_id: SupportsInt
    items: Sequence[Tuple[int, int]]

    def _get_response_patterns(self):
        return ["/b_setn", int(self.buffer_id)], None

    def to_osc(self) -> OscMessage:
        contents: List[int] = [int(self.buffer_id)]
        for index, count in self.items:
            contents.extend([int(index), int(count)])
        return OscMessage(RequestName.BUFFER_GET_CONTIGUOUS, *contents)


@dataclasses.dataclass
class GetControlBus(Request):
    """
    A ``/c_get`` request.

    ::

        >>> from supriya.contexts.requests import GetControlBus
        >>> request = GetControlBus(bus_ids=[1, 2, 3])
        >>> request.to_osc()
        OscMessage('/c_get', 1, 2, 3)
    """

    bus_ids: Sequence[SupportsInt]

    def _get_response_patterns(self):
        return ["/c_set", int(self.bus_ids[0])], None

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.CONTROL_BUS_GET, *(int(bus_id) for bus_id in self.bus_ids)
        )


@dataclasses.dataclass
class GetControlBusRange(Request):
    """
    A ``/c_getn`` request.

    ::

        >>> from supriya.contexts.requests import GetControlBusRange
        >>> request = GetControlBusRange(items=[(0, 4), (8, 16)])
        >>> request.to_osc()
        OscMessage('/c_getn', 0, 4, 8, 16)
    """

    items: Sequence[Tuple[SupportsInt, int]]

    def _get_response_patterns(self):
        return ["/c_setn", int(self.items[0][0])], None

    def to_osc(self) -> OscMessage:
        contents: List[int] = []
        for index, count in self.items:
            contents.extend([int(index), int(count)])
        return OscMessage(RequestName.CONTROL_BUS_GET_CONTIGUOUS, *contents)


@dataclasses.dataclass
class GetSynthControl(Request):
    """
    A ``/s_get`` request.

    ::

        >>> from supriya.contexts.requests import GetSynthControl
        >>> request = GetSynthControl(
        ...     synth_id=1000,
        ...     controls=["frequency", "amplitude"],
        ... )
        >>> request.to_osc()
        OscMessage('/s_get', 1000, 'frequency', 'amplitude')
    """

    synth_id: SupportsInt
    controls: Sequence[Union[int, str]]

    def _get_response_patterns(self):
        return ["/n_set", int(self.synth_id)], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[int, str]] = [int(self.synth_id)]
        for control in self.controls:
            contents.append(control if isinstance(control, str) else int(control))
        return OscMessage(RequestName.SYNTH_GET, *contents)


@dataclasses.dataclass
class GetSynthControlRange(Request):
    """
    A ``/s_getn`` request.

    ::

        >>> from supriya.contexts.requests import GetSynthControlRange
        >>> request = GetSynthControlRange(
        ...     synth_id=1000,
        ...     items=[("frequency", 8)],
        ... )
        >>> request.to_osc()
        OscMessage('/s_getn', 1000, 'frequency', 8)
    """

    synth_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], int]]

    def _get_response_patterns(self):
        return ["/n_setn", int(self.synth_id), self.items[0][0]], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[int, str]] = [int(self.synth_id)]
        for control, count in self.items:
            contents.extend(
                [control if isinstance(control, str) else int(control), int(count)]
            )
        return OscMessage(RequestName.SYNTH_GET_CONTIGUOUS, *contents)


@dataclasses.dataclass
class LoadSynthDefs(Request):
    """
    A ``/d_load`` request.

    ::

        >>> from supriya.contexts.requests import LoadSynthDefs, NewGroup
        >>> request = LoadSynthDefs(
        ...     path="path/to/synthdef-*.scsyndef",
        ...     on_completion=NewGroup(items=[(1000, "add to tail", 1)]),
        ... )
        >>> request.to_osc()
        OscMessage('/d_load', 'path/to/synthdef-*.scsyndef', OscMessage('/g_new', 1000, 1, 1))
    """

    path: PathLike
    on_completion: Optional[Requestable] = None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, str]] = [str(self.path)]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.SYNTHDEF_LOAD, *contents)


@dataclasses.dataclass
class LoadSynthDefDirectory(Request):
    """
    A ``/d_loadDir`` request.

    ::

        >>> from supriya.contexts.requests import LoadSynthDefDirectory, NewGroup
        >>> request = LoadSynthDefDirectory(
        ...     path="path/to/synthdefs/",
        ...     on_completion=NewGroup(items=[(1000, "add to tail", 1)]),
        ... )
        >>> request.to_osc()
        OscMessage('/d_loadDir', 'path/to/synthdefs/', OscMessage('/g_new', 1000, 1, 1))
    """

    path: PathLike
    on_completion: Optional[Requestable] = None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, str]] = [str(self.path)]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.SYNTHDEF_LOAD_DIR, *contents)


@dataclasses.dataclass
class MapAudioBusToNode(Request):
    """
    A ``/n_mapa`` request.

    ::

        >>> from supriya.contexts.requests import MapAudioBusToNode
        >>> request = MapAudioBusToNode(
        ...     node_id=1000,
        ...     items=[("frequency", 0)],
        ... )
        >>> request.to_osc()
        OscMessage('/n_mapa', 1000, 'frequency', 0)
    """

    node_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], SupportsInt]]

    def to_osc(self) -> OscMessage:
        contents: List[Union[int, str]] = [int(self.node_id)]
        for index_or_name, bus_index in self.items:
            contents.extend(
                [
                    index_or_name
                    if isinstance(index_or_name, str)
                    else int(index_or_name),
                    int(bus_index),
                ]
            )
        return OscMessage(RequestName.NODE_MAP_TO_AUDIO_BUS, *contents)


@dataclasses.dataclass
class MapAudioBusRangeToNode(Request):
    """
    A ``/n_mapan`` request.

    ::

        >>> from supriya.contexts.requests import MapAudioBusRangeToNode
        >>> request = MapAudioBusRangeToNode(
        ...     node_id=1000,
        ...     items=[("frequency", 0, 4)]
        ... )
        >>> request.to_osc()
        OscMessage('/n_mapan', 1000, 'frequency', 0, 4)
    """

    node_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], SupportsInt, int]]

    def to_osc(self) -> OscMessage:
        contents: List[Union[int, str]] = [int(self.node_id)]
        for index_or_name, bus_index, count in self.items:
            contents.extend(
                [
                    index_or_name
                    if isinstance(index_or_name, str)
                    else int(index_or_name),
                    int(bus_index),
                    int(count),
                ]
            )
        return OscMessage(RequestName.NODE_MAP_TO_AUDIO_BUS_CONTIGUOUS, *contents)


@dataclasses.dataclass
class MapControlBusToNode(Request):
    """
    A ``/n_map`` request.

    ::

        >>> from supriya.contexts.requests import MapControlBusToNode
        >>> request = MapControlBusToNode(
        ...     node_id=1000,
        ...     items=[("frequency", 0)],
        ... )
        >>> request.to_osc()
        OscMessage('/n_map', 1000, 'frequency', 0)
    """

    node_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], SupportsInt]]

    def to_osc(self) -> OscMessage:
        contents: List[Union[int, str]] = [int(self.node_id)]
        for index_or_name, bus_index in self.items:
            contents.extend(
                [
                    index_or_name
                    if isinstance(index_or_name, str)
                    else int(index_or_name),
                    int(bus_index),
                ]
            )
        return OscMessage(RequestName.NODE_MAP_TO_CONTROL_BUS, *contents)


@dataclasses.dataclass
class MapControlBusRangeToNode(Request):
    """
    A ``/n_mapn`` request.

    ::

        >>> from supriya.contexts.requests import MapControlBusRangeToNode
        >>> request = MapControlBusRangeToNode(
        ...     node_id=1000,
        ...     items=[("frequency", 0, 4)]
        ... )
        >>> request.to_osc()
        OscMessage('/n_mapn', 1000, 'frequency', 0, 4)
    """

    node_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], SupportsInt, int]]

    def to_osc(self) -> OscMessage:
        contents: List[Union[int, str]] = [int(self.node_id)]
        for index_or_name, bus_index, count in self.items:
            contents.extend(
                [
                    index_or_name
                    if isinstance(index_or_name, str)
                    else int(index_or_name),
                    int(bus_index),
                    int(count),
                ]
            )
        return OscMessage(RequestName.NODE_MAP_TO_CONTROL_BUS_CONTIGUOUS, *contents)


@dataclasses.dataclass
class MoveNodeAfter(Request):
    """
    A ``/n_after`` request.

    ::

        >>> from supriya.contexts.requests import MoveNodeAfter
        >>> request = MoveNodeAfter(
        ...     items=[(2000, 1001), (3000, 2001)],
        ... )
        >>> request.to_osc()
        OscMessage('/n_after', 2000, 1001, 3000, 2001)
    """

    items: Sequence[Tuple[SupportsInt, SupportsInt]]

    def to_osc(self) -> OscMessage:
        contents: List[int] = []
        for node_id, target_node_id in self.items:
            contents.extend([int(node_id), int(target_node_id)])
        return OscMessage(RequestName.NODE_AFTER, *contents)


@dataclasses.dataclass
class MoveNodeBefore(Request):
    """
    A ``/n_before`` request.

    ::

        >>> from supriya.contexts.requests import MoveNodeBefore
        >>> request = MoveNodeBefore(
        ...     items=[(2000, 1001), (3000, 2001)],
        ... )
        >>> request.to_osc()
        OscMessage('/n_before', 2000, 1001, 3000, 2001)
    """

    items: Sequence[Tuple[SupportsInt, SupportsInt]]

    def to_osc(self) -> OscMessage:
        contents: List[int] = []
        for node_id, target_node_id in self.items:
            contents.extend([int(node_id), int(target_node_id)])
        return OscMessage(RequestName.NODE_BEFORE, *contents)


@dataclasses.dataclass
class MoveNodeToGroupHead(Request):
    """
    A ``/g_head`` request.

    ::

        >>> from supriya.contexts.requests import MoveNodeToGroupHead
        >>> request = MoveNodeToGroupHead(
        ...     items=[(1000, 1), (1001, 1)],
        ... )
        >>> request.to_osc()
        OscMessage('/g_head', 1, 1000, 1, 1001)
    """

    items: Sequence[Tuple[SupportsInt, SupportsInt]]

    def to_osc(self) -> OscMessage:
        contents: List[int] = []
        for node_id, target_group_id in self.items:
            contents.extend([int(target_group_id), int(node_id)])
        return OscMessage(RequestName.GROUP_HEAD, *contents)


@dataclasses.dataclass
class MoveNodeToGroupTail(Request):
    """
    A ``/g_tail`` request.

    ::

        >>> from supriya.contexts.requests import MoveNodeToGroupTail
        >>> request = MoveNodeToGroupTail(
        ...     items=[(1000, 1), (1001, 1)],
        ... )
        >>> request.to_osc()
        OscMessage('/g_tail', 1, 1000, 1, 1001)
    """

    items: Sequence[Tuple[SupportsInt, SupportsInt]]

    def to_osc(self) -> OscMessage:
        contents: List[int] = []
        for node_id, target_group_id in self.items:
            contents.extend([int(target_group_id), int(node_id)])
        return OscMessage(RequestName.GROUP_TAIL, *contents)


@dataclasses.dataclass
class NewGroup(Request):
    """
    A ``/g_new`` request.

    ::

        >>> from supriya.contexts.requests import NewGroup
        >>> request = NewGroup(
        ...     items=[(1000, "ADD_TO_TAIL", 1)],
        ... )
        >>> request.to_osc()
        OscMessage('/g_new', 1000, 1, 1)
    """

    items: Sequence[Tuple[SupportsInt, AddActionLike, SupportsInt]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items: List[Tuple[SupportsInt, AddActionLike, SupportsInt]] = []
        for request in requests:
            if not isinstance(request, cls):
                continue
            items.extend(request.items)
        return [cls(items=items)]

    def to_osc(self) -> OscMessage:
        contents = []
        for group_id, add_action, target_node_id in self.items:
            contents.extend(
                [
                    int(group_id),
                    int(AddAction.from_expr(add_action)),
                    int(target_node_id),
                ]
            )
        return OscMessage(RequestName.GROUP_NEW, *contents)


@dataclasses.dataclass
class NewParallelGroup(Request):
    """
    A ``/p_new`` request.

    ::

        >>> from supriya.contexts.requests import NewParallelGroup
        >>> request = NewParallelGroup(
        ...     items=[(1000, "ADD_TO_TAIL", 1)]
        ... )
        >>> request.to_osc()
        OscMessage('/p_new', 1000, 1, 1)
    """

    items: Sequence[Tuple[SupportsInt, AddActionLike, SupportsInt]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items: List[Tuple[SupportsInt, AddActionLike, SupportsInt]] = []
        for request in requests:
            if not isinstance(request, cls):
                continue
            items.extend(request.items)
        return [cls(items=items)]

    def to_osc(self) -> OscMessage:
        contents = []
        for group_id, add_action, target_node_id in self.items:
            contents.extend(
                [
                    int(group_id),
                    int(AddAction.from_expr(add_action)),
                    int(target_node_id),
                ]
            )
        return OscMessage(RequestName.PARALLEL_GROUP_NEW, *contents)


@dataclasses.dataclass
class NewSynth(Request):
    """
    A ``/s_new`` request.

    ::

        >>> from supriya import default
        >>> from supriya.contexts.requests import NewSynth
        >>> request = NewSynth(
        ...     synthdef=default,
        ...     synth_id=1001,
        ...     add_action="ADD_TO_TAIL",
        ...     target_node_id=1000,
        ...     controls={
        ...         "frequency": 432.0,
        ...         "amplitude": 0.5,
        ...         "panning": "c0",
        ...     },
        ... )
        >>> request.to_osc()
        OscMessage('/s_new', 'default', 1001, 1, 1000, 'amplitude', 0.5, 'frequency', 432.0, 'panning', 'c0')
    """

    synthdef: Union[SynthDef, str]
    synth_id: SupportsInt
    add_action: AddActionLike
    target_node_id: SupportsInt
    controls: Optional[Dict[Union[int, str], Union[SupportsFloat, str]]] = None

    def to_osc(self) -> OscMessage:
        contents: List[Union[float, str]] = [
            self.synthdef.actual_name
            if isinstance(self.synthdef, SynthDef)
            else self.synthdef,
            int(self.synth_id),
            int(AddAction.from_expr(self.add_action)),
            int(self.target_node_id),
        ]
        for key, value in sorted((self.controls or {}).items()):
            contents.append(key if isinstance(key, str) else int(key))
            contents.append(value if isinstance(value, str) else float(value))
        return OscMessage(RequestName.SYNTH_NEW, *contents)


@dataclasses.dataclass
class NormalizeBuffer(Request):
    """
    A ``/b_gen`` ``normalize`` (or ``wnormalize``) request.

    ::

        >>> from supriya.contexts.requests import NormalizeBuffer
        >>> request = NormalizeBuffer(
        ...     buffer_id=1,
        ...     new_maximum=0.999,
        ... )
        >>> request.to_osc()
        OscMessage('/b_gen', 1, 'normalize', 0.999)
    """

    buffer_id: SupportsInt
    new_maximum: float = 1.0
    as_wavetable: bool = False

    def _get_response_patterns(self):
        return ["/done", "/b_gen", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.BUFFER_GENERATE,
            int(self.buffer_id),
            "wnormalize" if self.as_wavetable else "normalize",
            float(self.new_maximum),
        )


@dataclasses.dataclass
class OrderNodes(Request):
    """
    A ``/n_order`` request.

    ::

        >>> from supriya.contexts.requests import OrderNodes
        >>> request = OrderNodes(
        ...     add_action="ADD_AFTER",
        ...     target_node_id=1000,
        ...     node_ids=[1003, 1002, 1001],
        ... )
        >>> request.to_osc()
        OscMessage('/n_order', 3, 1000, 1003, 1002, 1001)
    """

    add_action: AddActionLike
    target_node_id: SupportsInt
    node_ids: Sequence[SupportsInt]

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.NODE_ORDER,
            int(AddAction.from_expr(self.add_action)),
            int(self.target_node_id),
            *(int(node_id) for node_id in self.node_ids),
        )


@dataclasses.dataclass
class QueryBuffer(Request):
    """
    A ``/b_query`` request.

    ::

        >>> from supriya.contexts.requests import QueryBuffer
        >>> request = QueryBuffer(buffer_ids=[1, 2, 3])
        >>> request.to_osc()
        OscMessage('/b_query', 1, 2, 3)
    """

    buffer_ids: Sequence[int]

    def _get_response_patterns(self):
        # TODO: We should be able to gather multiple responses
        return ["/b_info", self.buffer_ids[-1]], None

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.BUFFER_QUERY, *(int(buffer_id) for buffer_id in self.buffer_ids)
        )


@dataclasses.dataclass
class QueryNode(Request):
    """
    A ``/n_query`` request.

    ::

        >>> from supriya.contexts.requests import QueryNode
        >>> request = QueryNode(node_ids=[1000, 1001])
        >>> request.to_osc()
        OscMessage('/n_query', 1000, 1001)
    """

    node_ids: Sequence[SupportsInt]

    def _get_response_patterns(self):
        # TODO: We should be able to gather multiple responses
        return ["/n_info", int(self.node_ids[-1])], None

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.NODE_QUERY, *(int(node_id) for node_id in self.node_ids)
        )


@dataclasses.dataclass
class QueryStatus(Request):
    """
    A ``/status`` request.

    ::

        >>> from supriya.contexts.requests import QueryStatus
        >>> request = QueryStatus()
        >>> request.to_osc()
        OscMessage('/status')
    """

    def _get_response_patterns(self):
        return ["/status.reply"], None

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.STATUS)


@dataclasses.dataclass
class QueryTree(Request):
    """
    A ``/g_queryTree`` request.

    ::

        >>> from supriya.contexts.requests import QueryTree
        >>> request = QueryTree(items=[(0, True)])
        >>> request.to_osc()
        OscMessage('/g_queryTree', 0, 1)
    """

    items: Sequence[Tuple[SupportsInt, bool]]

    def _get_response_patterns(self):
        # TODO: We should be able to gather multiple responses
        return [
            "/g_queryTree.reply",
            int(self.items[-1][1]),
            int(self.items[-1][0]),
        ], None

    def to_osc(self) -> OscMessage:
        contents: List[int] = []
        for node_id, flag in self.items:
            contents.extend([int(node_id), int(bool(flag))])
        return OscMessage(RequestName.GROUP_QUERY_TREE, *contents)


@dataclasses.dataclass
class QueryVersion(Request):
    """
    A ``/version`` request.

    ::

        >>> from supriya.contexts.requests import QueryVersion
        >>> request = QueryVersion()
        >>> request.to_osc()
        OscMessage('/version')
    """

    def _get_response_patterns(self):
        return ["/version.reply"], None

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.VERSION)


@dataclasses.dataclass
class Quit(Request):
    """
    A ``/quit`` request.

    ::

        >>> from supriya.contexts.requests import Quit
        >>> request = Quit()
        >>> request.to_osc()
        OscMessage('/quit')
    """

    def _get_response_patterns(self):
        return ["/done", "/quit"], None

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.QUIT)


@dataclasses.dataclass
class ReadBuffer(Request):
    """
    A ``/b_read`` request.

    ::

        >>> from supriya.contexts.requests import NormalizeBuffer, ReadBuffer
        >>> request = ReadBuffer(
        ...     buffer_id=1,
        ...     path="path/to/audio.wav",
        ...     on_completion=NormalizeBuffer(buffer_id=1),
        ... )
        >>> request.to_osc()
        OscMessage('/b_read', 1, 'path/to/audio.wav', 0, -1, 0, 0, OscMessage('/b_gen', 1, 'normalize', 1.0))
    """

    buffer_id: SupportsInt
    path: PathLike
    frame_count: int = -1
    leave_open: bool = False
    starting_frame_in_buffer: int = 0
    starting_frame_in_file: int = 0
    on_completion: Optional[Requestable] = None

    def _get_response_patterns(self):
        return ["/done", "/b_read", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int, str]] = [
            int(self.buffer_id),
            str(self.path),
            int(self.starting_frame_in_file),
            int(self.frame_count),
            int(self.starting_frame_in_buffer),
            int(bool(self.leave_open)),
        ]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_READ, *contents)


@dataclasses.dataclass
class ReadBufferChannel(Request):
    """
    A ``/b_readChannel`` request.

    ::

        >>> from supriya.contexts.requests import NormalizeBuffer, ReadBufferChannel
        >>> request = ReadBufferChannel(
        ...     buffer_id=1,
        ...     path="path/to/audio.wav",
        ...     channel_indices=[2, 4],
        ...     on_completion=NormalizeBuffer(buffer_id=1),
        ... )
        >>> request.to_osc()
        OscMessage('/b_readChannel', 1, 'path/to/audio.wav', 0, -1, 0, 0, 2, 4, OscMessage('/b_gen', 1, 'normalize', 1.0))
    """

    buffer_id: SupportsInt
    path: PathLike
    channel_indices: Optional[Sequence[int]] = None
    frame_count: int = -1
    leave_open: bool = False
    starting_frame_in_buffer: int = 0
    starting_frame_in_file: int = 0
    on_completion: Optional[Requestable] = None

    def _get_response_patterns(self):
        return ["/done", "/b_readChannel", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int, str]] = [
            int(self.buffer_id),
            str(self.path),
            int(self.starting_frame_in_file),
            int(self.frame_count),
            int(self.starting_frame_in_buffer),
            int(bool(self.leave_open)),
            *(int(channel_index) for channel_index in self.channel_indices or []),
        ]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_READ_CHANNEL, *contents)


@dataclasses.dataclass
class ReceiveSynthDefs(Request):
    r"""
    A ``/d_recv`` request.

    ::

        >>> from supriya import default
        >>> from supriya.contexts.requests import NewSynth, ReceiveSynthDefs
        >>> request = ReceiveSynthDefs(
        ...     synthdefs=[default],
        ...     on_completion=NewSynth(
        ...         synthdef=default,
        ...         synth_id=1000,
        ...         add_action="ADD_TO_TAIL",
        ...         target_node_id=1,
        ...     ),
        ... )
        >>> request.to_osc()
        OscMessage('/d_recv', b'SCgf\x00\x00\x00\x02\x00\x01\x07default\x00\x00\x00\x0c\x00\x00\x00\x00>\x99\x99\x9a<#\xd7\n?333@\x00\x00\x00\xbe\xcc\xcc\xcd>\xcc\xcc\xcdEz\x00\x00E\x9c@\x00E\x1c@\x00EH\x00\x00?\x80\x00\x00\x00\x00\x00\x05\x00\x00\x00\x00=\xcc\xcc\xcdC\xdc\x00\x00?\x80\x00\x00?\x00\x00\x00\x00\x00\x00\x05\tamplitude\x00\x00\x00\x01\tfrequency\x00\x00\x00\x02\x04gate\x00\x00\x00\x03\x03out\x00\x00\x00\x00\x03pan\x00\x00\x00\x04\x00\x00\x00\x14\x07Control\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x07Control\x01\x00\x00\x00\x00\x00\x00\x00\x04\x00\x01\x01\x01\x01\x01\x06VarSaw\x02\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\x02\x05Linen\x01\x00\x00\x00\x05\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x02\xff\xff\xff\xff\x00\x00\x00\x02\xff\xff\xff\xff\x00\x00\x00\x03\xff\xff\xff\xff\x00\x00\x00\x01\xff\xff\xff\xff\x00\x00\x00\x04\x01\x04Rand\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\xff\xff\xff\xff\x00\x00\x00\x05\xff\xff\xff\xff\x00\x00\x00\x00\x00\x0cBinaryOpUGen\x01\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x04\x00\x00\x00\x00\x01\x06VarSaw\x02\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\x00\x00\x05\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\x02\x04Rand\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x06\x00\x0cBinaryOpUGen\x01\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00\x07\x00\x00\x00\x00\x01\x06VarSaw\x02\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\x00\x00\x08\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\x02\x04Sum3\x02\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\t\x00\x00\x00\x00\x02\x0cBinaryOpUGen\x02\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02\x00\x00\x00\n\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x01\x02\x04Rand\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\xff\xff\xff\xff\x00\x00\x00\x07\xff\xff\xff\xff\x00\x00\x00\x08\x00\x04Rand\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\xff\xff\xff\xff\x00\x00\x00\t\xff\xff\xff\xff\x00\x00\x00\n\x00\x05XLine\x01\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00\x00\x0c\x00\x00\x00\x00\x00\x00\x00\r\x00\x00\x00\x00\xff\xff\xff\xff\x00\x00\x00\x0b\xff\xff\xff\xff\x00\x00\x00\x00\x01\x03LPF\x02\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x00\x00\x0b\x00\x00\x00\x00\x00\x00\x00\x0e\x00\x00\x00\x00\x02\x0cBinaryOpUGen\x02\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02\x00\x00\x00\x0f\x00\x00\x00\x00\x00\x00\x00\x03\x00\x00\x00\x00\x02\x0cBinaryOpUGen\x02\x00\x00\x00\x02\x00\x00\x00\x01\x00\x02\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x02\x04Pan2\x02\x00\x00\x00\x03\x00\x00\x00\x02\x00\x00\x00\x00\x00\x11\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x03\xff\xff\xff\xff\x00\x00\x00\x0b\x02\x02\tOffsetOut\x02\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x12\x00\x00\x00\x00\x00\x00\x00\x12\x00\x00\x00\x01\x00\x00', OscMessage('/s_new', 'default', 1000, 1, 1))
    """

    synthdefs: Sequence[SynthDef]
    on_completion: Optional[Requestable] = None

    def to_osc(self) -> OscMessage:
        contents = [SynthDefCompiler.compile_synthdefs(self.synthdefs)]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.SYNTHDEF_RECEIVE, *contents)


@dataclasses.dataclass
class ReleaseNode(Request):
    node_id: SupportsInt
    has_gate: bool = False
    force: bool = False

    def to_osc(self) -> OscMessage:
        if not self.force and self.has_gate:
            return SetNodeControl(node_id=self.node_id, items=[("gate", 0)]).to_osc()
        return FreeNode(node_ids=[self.node_id]).to_osc()


@dataclasses.dataclass
class RunNode(Request):
    """
    A ``/n_run`` request.

    ::

        >>> from supriya.contexts.requests import RunNode
        >>> request = RunNode(items=[(1000, True)])
        >>> request.to_osc()
        OscMessage('/n_run', 1000, 1)
    """

    items: Sequence[Tuple[SupportsInt, bool]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items: List[Tuple[SupportsInt, bool]] = []
        for request in requests:
            if not isinstance(request, cls):
                continue
            items.extend(request.items)
        return [cls(items=items)]

    def to_osc(self) -> OscMessage:
        contents: List[int] = []
        for node_id, flag in self.items:
            contents.extend([int(node_id), int(flag)])
        return OscMessage(RequestName.NODE_RUN, *contents)


@dataclasses.dataclass
class SetBuffer(Request):
    """
    A ``/b_set`` request.

    ::

        >>> from supriya.contexts.requests import SetBuffer
        >>> request = SetBuffer(
        ...     buffer_id=1,
        ...     items=[(0, 0.5), (8, 0.25)],
        ... )
        >>> request.to_osc()
        OscMessage('/b_set', 1, 0, 0.5, 8, 0.25)
    """

    buffer_id: SupportsInt
    items: Sequence[Tuple[int, float]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items_by_buffer_id: Dict[int, List[Tuple[int, float]]] = {}
        for request in requests:
            if isinstance(request, cls):
                items_by_buffer_id.setdefault(int(request.buffer_id), []).extend(
                    request.items
                )
        return [
            cls(buffer_id=buffer_id, items=items)
            for buffer_id, items in sorted(items_by_buffer_id.items())
        ]

    def to_osc(self) -> OscMessage:
        contents: List[float] = [int(self.buffer_id)]
        for index, value in self.items:
            contents.extend([int(index), float(value)])
        return OscMessage(RequestName.BUFFER_SET, *contents)


@dataclasses.dataclass
class SetBufferRange(Request):
    """
    A ``/b_setn`` request.

    ::

        >>> from supriya.contexts.requests import SetBufferRange
        >>> request = SetBufferRange(
        ...     buffer_id=1,
        ...     items=[(0, (0.1, 0.2, 0.3)), (8, (0.4, 0.5))],
        ... )
        >>> request.to_osc()
        OscMessage('/b_setn', 1, 0, 3, 0.1, 0.2, 0.3, 8, 2, 0.4, 0.5)
    """

    buffer_id: SupportsInt
    items: Sequence[Tuple[int, Sequence[float]]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items_by_buffer_id: Dict[int, List[Tuple[int, Sequence[float]]]] = {}
        for request in requests:
            if isinstance(request, cls):
                items_by_buffer_id.setdefault(int(request.buffer_id), []).extend(
                    request.items
                )
        return [
            cls(buffer_id=buffer_id, items=items)
            for buffer_id, items in sorted(items_by_buffer_id.items())
        ]

    def to_osc(self) -> OscMessage:
        contents: List[float] = [int(self.buffer_id)]
        for index, values in self.items:
            contents.extend(
                [int(index), len(values), *(float(value) for value in values)]
            )
        return OscMessage(RequestName.BUFFER_SET_CONTIGUOUS, *contents)


@dataclasses.dataclass
class SetControlBus(Request):
    """
    A ``/c_set`` request.

    ::

        >>> from supriya.contexts.requests import SetControlBus
        >>> request = SetControlBus(items=[(0, 0.5), (4, 0.75)])
        >>> request.to_osc()
        OscMessage('/c_set', 0, 0.5, 4, 0.75)
    """

    items: Sequence[Tuple[SupportsInt, float]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items: List[Tuple[SupportsInt, float]] = []
        for request in requests:
            if not isinstance(request, cls):
                continue
            items.extend(request.items)
        return [cls(items=items)]

    def to_osc(self) -> OscMessage:
        contents: List[float] = []
        for index, value in self.items:
            contents.extend([int(index), float(value)])
        return OscMessage(RequestName.CONTROL_BUS_SET, *contents)


@dataclasses.dataclass
class SetControlBusRange(Request):
    """
    A ``/c_setn`` request.

    ::

        >>> from supriya.contexts.requests import SetControlBusRange
        >>> request = SetControlBusRange(
        ...     items=[(8, [1.1, 2.2, 3.3]), (16, [4.4, 5.5])],
        ... )
        >>> request.to_osc()
        OscMessage('/c_setn', 8, 3, 1.1, 2.2, 3.3, 16, 2, 4.4, 5.5)
    """

    items: Sequence[Tuple[SupportsInt, Sequence[float]]]

    @classmethod
    def merge(cls, requests: List["Request"]) -> List["Request"]:
        items: List[Tuple[SupportsInt, Sequence[float]]] = []
        for request in requests:
            if not isinstance(request, cls):
                continue
            items.extend(request.items)
        return [cls(items=items)]

    def to_osc(self) -> OscMessage:
        contents: List[float] = []
        for index, values in self.items:
            contents.extend(
                [int(index), len(values), *(float(value) for value in values)]
            )
        return OscMessage(RequestName.CONTROL_BUS_SET_CONTIGUOUS, *contents)


@dataclasses.dataclass
class SetNodeControl(Request):
    """
    A ``/n_set`` request.

    ::

        >>> from supriya.contexts.requests import SetNodeControl
        >>> request = SetNodeControl(
        ...     node_id=1000,
        ...     items=[
        ...         ("frequency", 440.0),
        ...         ("amplitude", 1.0),
        ...         (3, 1.234),
        ...         ("positions", [0.5, 0.25, 0.75]),
        ...         (4, [0.1, 0.2]),
        ...     ],
        ... )
        >>> request.to_osc()
        OscMessage('/n_set', 1000, 'frequency', 440.0, 'amplitude', 1.0, 3, 1.234, 'positions', [0.5, 0.25, 0.75], 4, [0.1, 0.2])
    """

    node_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], Union[float, Sequence[float]]]]

    def to_osc(self) -> OscMessage:
        contents: List[Union[float, str, List[float]]] = [int(self.node_id)]
        for control, values in self.items:
            contents.append(control if isinstance(control, str) else int(control))
            if isinstance(values, Sequence):
                contents.append([float(value) for value in values])
            else:
                contents.append(float(values))
        return OscMessage(RequestName.NODE_SET, *contents)


@dataclasses.dataclass
class SetNodeControlRange(Request):
    """
    A ``/n_setn`` request.

    ::

        >>> from supriya.contexts.requests import SetNodeControlRange
        >>> request = SetNodeControlRange(
        ...     node_id=1000,
        ...     items=[
        ...         ("frequency", (440.0, 441.0, 432.0)),
        ...         (3, (0.5, 0.25, 0.75)),
        ...     ],
        ... )
        >>> request.to_osc()
        OscMessage('/n_setn', 1000, 'frequency', 3, 440.0, 441.0, 432.0, 3, 3, 0.5, 0.25, 0.75)
    """

    node_id: SupportsInt
    items: Sequence[Tuple[Union[int, str], Sequence[float]]]

    def to_osc(self) -> OscMessage:
        contents: List[Union[float, str]] = [int(self.node_id)]
        for control, values in self.items:
            contents.extend(
                [
                    control if isinstance(control, str) else int(control),
                    len(values),
                    *(float(value) for value in values),
                ]
            )
        return OscMessage(RequestName.NODE_SET_CONTIGUOUS, *contents)


@dataclasses.dataclass
class Sync(Request):
    """
    A ``/sync`` request.

    ::

        >>> from supriya.contexts.requests import Sync
        >>> request = Sync(42)
        >>> request.to_osc()
        OscMessage('/sync', 42)
    """

    sync_id: int

    def _get_response_patterns(self):
        return ["/synced", self.sync_id], None

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.SYNC, int(self.sync_id))


@dataclasses.dataclass
class ToggleErrorReporting(Request):
    """
    A ``/error`` request.

    ::

        >>> from supriya.contexts.requests import ToggleErrorReporting
        >>> request = ToggleErrorReporting(code=1)
        >>> request.to_osc()
        OscMessage('/error', 1)
    """

    code: Literal[0, 1, -1, -2]

    def to_osc(self) -> OscMessage:
        return OscMessage(RequestName.ERROR, int(self.code))


@dataclasses.dataclass
class ToggleNotifications(Request):
    """
    A ``/notify`` request.

    ::

        >>> from supriya.contexts.requests import ToggleNotifications
        >>> request = ToggleNotifications(True)
        >>> request.to_osc()
        OscMessage('/notify', 1)
    """

    should_notify: bool
    client_id: Optional[int] = None

    def _get_response_patterns(self):
        return ["/done", "/notify"], ["/fail", "/notify"]

    def to_osc(self) -> OscMessage:
        contents: List[int] = [int(bool(self.should_notify))]
        if self.client_id is not None:
            contents.append(int(self.client_id))
        return OscMessage(RequestName.NOTIFY, *contents)


@dataclasses.dataclass
class TraceNode(Request):
    """
    A ``/n_trace`` request.

    ::

        >>> from supriya.contexts.requests import TraceNode
        >>> request = TraceNode(node_ids=[1, 2, 3])
        >>> request.to_osc()
        OscMessage('/n_trace', 1, 2, 3)
    """

    node_ids: Sequence[SupportsInt]

    def to_osc(self) -> OscMessage:
        return OscMessage(
            RequestName.NODE_TRACE, *(int(node_id) for node_id in self.node_ids)
        )


@dataclasses.dataclass
class WriteBuffer(Request):
    """
    A ``/b_write`` request.

    ::

        >>> from supriya.contexts.requests import FreeBuffer, WriteBuffer
        >>> request = WriteBuffer(
        ...     buffer_id=1,
        ...     path="path/to/audio.wav",
        ...     header_format="wav",
        ...     sample_format="int24",
        ...     on_completion=FreeBuffer(buffer_id=1)
        ... )
        >>> request.to_osc()
        OscMessage('/b_write', 1, 'path/to/audio.wav', 'wav', 'int24', -1, 0, 0, OscMessage('/b_free', 1))
    """

    buffer_id: SupportsInt
    path: PathLike
    header_format: HeaderFormatLike
    sample_format: SampleFormatLike
    frame_count: int = -1
    starting_frame: int = 0
    leave_open: bool = False
    on_completion: Optional[Requestable] = None

    def _get_response_patterns(self):
        return ["/done", "/b_write", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, int, str]] = [
            int(self.buffer_id),
            str(self.path),
            HeaderFormat.from_expr(self.header_format).name.lower(),
            SampleFormat.from_expr(self.sample_format).name.lower(),
            int(self.frame_count),
            int(self.starting_frame),
            int(bool(self.leave_open)),
        ]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_WRITE, *contents)


@dataclasses.dataclass
class ZeroBuffer(Request):
    """
    A ``/b_zero`` request.

    ::

        >>> from supriya.contexts.requests import FreeNode, ZeroBuffer
        >>> request = ZeroBuffer(
        ...     buffer_id=1,
        ...     on_completion=FreeNode(node_ids=[1000]),
        ... )
        >>> request.to_osc()
        OscMessage('/b_zero', 1, OscMessage('/n_free', 1000))
    """

    buffer_id: SupportsInt
    on_completion: Optional[Requestable] = None

    def _get_response_patterns(self):
        return ["/done", "/b_zero", self.buffer_id], None

    def to_osc(self) -> OscMessage:
        contents: List[Union[OscBundle, OscMessage, SupportsInt]] = [
            int(self.buffer_id)
        ]
        if self.on_completion:
            contents.append(self.on_completion.to_osc())
        return OscMessage(RequestName.BUFFER_ZERO, *contents)



from .bases import UGen, param, ugen


@ugen(ar=True, kr=True)
class Gendy1(UGen):
    """
    A dynamic stochastic synthesis generator.

    ::

        >>> gendy_1 = supriya.ugens.Gendy1.ar(
        ...     adparam=1,
        ...     ampdist=1,
        ...     ampscale=0.5,
        ...     ddparam=1,
        ...     durdist=1,
        ...     durscale=0.5,
        ...     init_cps=12,
        ...     knum=10,
        ...     maxfrequency=660,
        ...     minfrequency=440,
        ... )
        >>> gendy_1
        Gendy1.ar()
    """

    ### CLASS VARIABLES ###

    ampdist = param(1)
    durdist = param(1)
    adparam = param(1)
    ddparam = param(1)
    minfrequency = param(440)
    maxfrequency = param(660)
    ampscale = param(0.5)
    durscale = param(0.5)
    init_cps = param(12)
    knum = param(None)

    ### INITIALIZER ###

    def __init__(
        self,
        calculation_rate=None,
        adparam=1,
        ampdist=1,
        ampscale=0.5,
        ddparam=1,
        durdist=1,
        durscale=0.5,
        init_cps=12,
        knum=None,
        maxfrequency=660,
        minfrequency=440,
    ):
        if knum is None:
            knum = init_cps
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            adparam=adparam,
            ampdist=ampdist,
            ampscale=ampscale,
            ddparam=ddparam,
            durdist=durdist,
            durscale=durscale,
            init_cps=init_cps,
            knum=knum,
            maxfrequency=maxfrequency,
            minfrequency=minfrequency,
        )


@ugen(ar=True, kr=True)
class Gendy2(UGen):
    """
    A dynamic stochastic synthesis generator.

    ::

        >>> gendy_2 = supriya.ugens.Gendy2.ar(
        ...     a=1.17,
        ...     adparam=1,
        ...     ampdist=1,
        ...     ampscale=0.5,
        ...     c=0.31,
        ...     ddparam=1,
        ...     durdist=1,
        ...     durscale=0.5,
        ...     init_cps=12,
        ...     knum=10,
        ...     maxfrequency=660,
        ...     minfrequency=440,
        ... )
        >>> gendy_2
        Gendy2.ar()
    """

    ampdist = param(1)
    durdist = param(1)
    adparam = param(1)
    ddparam = param(1)
    minfrequency = param(440)
    maxfrequency = param(660)
    ampscale = param(0.5)
    durscale = param(0.5)
    init_cps = param(12)
    knum = param(None)
    a = param(1.17)
    c = param(0.31)


@ugen(ar=True, kr=True)
class Gendy3(UGen):
    """
    A dynamic stochastic synthesis generator.

    ::

        >>> gendy_3 = supriya.ugens.Gendy3.ar(
        ...     adparam=1,
        ...     ampdist=1,
        ...     ampscale=0.5,
        ...     ddparam=1,
        ...     durdist=1,
        ...     durscale=0.5,
        ...     frequency=440,
        ...     init_cps=12,
        ...     knum=10,
        ... )
        >>> gendy_3
        Gendy3.ar()
    """

    ampdist = param(1)
    durdist = param(1)
    adparam = param(1)
    ddparam = param(1)
    frequency = param(440)
    ampscale = param(0.5)
    durscale = param(0.5)
    init_cps = param(12)
    knum = param(None)



import platform

import pytest


@pytest.mark.skipif(
    platform.system() == "Windows", reason="SHM not built under Windows"
)
def test_shared_memory(server):
    from supriya.contexts import shm

    shared_memory = shm.ServerSHM(
        server.options.port, server.options.control_bus_channel_count
    )
    bus = server.add_bus()
    assert bus.get() == shared_memory[int(bus)] == 0.0
    for value in [1.0, 23.0, 666.0, 0.5]:
        bus.set(value)
        assert bus.get() == shared_memory[int(bus)] == value
    assert shared_memory[:2] == [0.5, 0]



from queue import Empty, PriorityQueue, Queue
from threading import RLock
from typing import (
    Callable,
    Coroutine,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)
from uuid import UUID, uuid4

from ..clocks import BaseClock, CallbackEvent, Clock, ClockContext, OfflineClock
from ..contexts import Context, ContextObject, Node
from .eventpatterns import Pattern
from .events import Event, Priority, StartEvent, StopEvent


class PatternPlayer:
    """
    A pattern player.

    Coordinates interactions between a pattern, a clock_context, and a clock.
    """

    def __init__(
        self,
        pattern: Pattern,
        context: Context,
        clock: BaseClock,
        callback: Optional[
            Callable[
                ["PatternPlayer", ClockContext, Event, Priority], Optional[Coroutine]
            ]
        ] = None,
        target_node: Optional[Node] = None,
        uuid: Optional[UUID] = None,
    ) -> None:
        self._pattern = pattern
        self._context = context
        self._clock = clock
        self._callback = callback
        self._lock = RLock()
        self._queue: Queue[
            Tuple[float, Priority, Union[int, Tuple[int, int]], Optional[Event]]
        ] = PriorityQueue()
        self._is_running = False
        self._is_stopping = False
        self._proxies_by_uuid: Dict[Union[UUID, Tuple[UUID, int]], ContextObject] = {}
        self._notes_by_uuid: Dict[Union[UUID, Tuple[UUID, int]], float] = {}
        self._uuid: UUID = uuid or uuid4()
        self._target_node = target_node
        self._next_delta: Optional[float] = None
        self._initial_seconds: Optional[float] = None

    def _clock_callback(
        self, clock_context: ClockContext, *args, **kwargs
    ) -> Optional[float]:
        for clock_context, seconds, offset, events in self._find_events(clock_context):
            if self._initial_seconds is None:
                self._initial_seconds = seconds
            with self._context.at(seconds):
                for event, priority in events:
                    event.perform(
                        self._context,
                        self._proxies_by_uuid,
                        current_offset=offset,
                        notes_mapping=self._notes_by_uuid,
                        priority=priority,
                        target_node=self._target_node,
                    )
                    if self._callback is not None:
                        self._callback(self, clock_context, event, priority)
        return self._next_delta

    def _find_events(
        self, clock_context: ClockContext
    ) -> Iterable[Tuple[ClockContext, float, float, Sequence[Tuple[Event, Priority]]]]:
        with self._lock:
            current_offset = float("-inf")
            events: List[Tuple[Event, Priority]] = []
            if (
                not cast(CallbackEvent, clock_context.event).invocations
                and self._callback
            ):
                yield clock_context, clock_context.desired_moment.seconds, clock_context.desired_moment.offset, [
                    (StartEvent(), Priority.START)
                ]
            while True:
                try:
                    offset, priority, index, event = self._queue.get(block=False)
                except Empty:
                    if events:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, events
                    self._is_running = False
                    if self._callback:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, [
                            (StopEvent(), Priority.STOP)
                        ]
                    self._next_delta = None
                    return
                except Exception:
                    self._next_delta = None
                    return
                if offset == float("-inf"):
                    offset = clock_context.desired_moment.offset
                delta = offset - clock_context.desired_moment.offset
                if delta:
                    self._queue.put((offset, priority, index, event))
                    if events:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, events
                    self._next_delta = delta
                    return
                if not isinstance(event, Event):
                    if self._consume_iterator(offset):
                        if self._callback:
                            yield clock_context, clock_context.desired_moment.seconds, current_offset, [
                                (StopEvent(), Priority.STOP)
                            ]
                        self._next_delta = None
                        return
                elif offset != current_offset:
                    if events:
                        yield clock_context, clock_context.desired_moment.seconds, current_offset, events
                    current_offset = offset
                    events = [(event, priority)]
                else:
                    events.append((event, priority))

    def _consume_iterator(self, current_offset: float) -> bool:
        try:
            try:
                index, consumed_event = self._iterator.send(self._is_stopping)
            except TypeError:
                if self._is_stopping:
                    return True
                index, consumed_event = next(self._iterator)
            for subindex, (expanded_offset, priority, expanded_event) in enumerate(
                consumed_event.expand(current_offset)
            ):
                self._queue.put(
                    (expanded_offset, priority, (index, subindex), expanded_event)
                )
            self._queue.put(
                (
                    float(current_offset + consumed_event.delta),
                    Priority.NONE,
                    (index, 0),
                    None,
                )
            )
        except StopIteration:
            pass
        return False

    def _stop_callback(
        self, clock_context: ClockContext, *args, **kwargs
    ) -> Optional[float]:
        with self._lock:
            # Do we need to rebuild the queue? Yes.
            # Do we need to free all playing notes? Yes.
            # How do we handle when there are already stop events in the queue?
            # They'll be no-ops when performed.
            self._is_stopping = True
            self._clock.reschedule(
                self._clock_event_id, schedule_at=clock_context.desired_moment.offset
            )
            self._reschedule_queue(clock_context.desired_moment.offset)
            self._free_all_notes(clock_context.desired_moment.seconds)
        return None

    def _enumerate(
        self, iterator: Generator[Event, bool, None]
    ) -> Generator[Tuple[int, Event], bool, None]:
        index = 0
        should_stop = False
        while True:
            try:
                should_stop = yield (index, iterator.send(should_stop)) or should_stop
            except TypeError:
                should_stop = yield (index, next(iterator))
            except StopIteration:
                return
            index += 1

    def _free_all_notes(self, current_seconds: float) -> None:
        if not self._notes_by_uuid:
            return
        with self._context.at(current_seconds):
            while self._notes_by_uuid:
                uuid, _ = self._notes_by_uuid.popitem()
                if not isinstance(node := self._proxies_by_uuid.pop(uuid), Node):
                    raise RuntimeError
                self._context.free_node(node)

    def _reschedule_queue(self, current_offset: float) -> None:
        events: List[
            Tuple[float, Priority, Union[int, Tuple[int, int]], Optional[Event]]
        ] = []
        while not self._queue.empty():
            events.append(self._queue.get())
        if not events:
            return
        delta = events[0][0] - current_offset
        for event in events:
            offset, priority, index, event_ = event
            self._queue.put((offset - delta, priority, index, event_))

    def play(
        self,
        quantization: Optional[str] = None,
        at: Optional[float] = None,
        until: Optional[float] = None,
    ) -> None:
        with self._lock:
            if self._is_running:
                return
            self._iterator = self._enumerate(iter(self._pattern))
            self._queue.put((float("-inf"), Priority.NONE, (0, 0), None))
            self._is_running = True
            self._is_stopping = False
        self._clock_event_id = self._clock.cue(
            self._clock_callback,
            event_type=3,
            quantization=quantization,
        )
        if until:
            self._clock.schedule(self._stop_callback, event_type=2, schedule_at=until)
        if (
            isinstance(self._clock, (Clock, OfflineClock))
            and not self._clock.is_running
        ):
            self._clock.start(initial_time=at)

    def stop(self, quantization: Optional[str] = None) -> None:
        with self._lock:
            if not self._is_running or self._is_stopping:
                return
            self._clock.cue(
                self._stop_callback, event_type=2, quantization=quantization
            )

    def uuid_to_note_id(self, uuid: UUID, index: Optional[int] = None) -> float:
        if index is not None:
            return self._notes_by_uuid[uuid, index]
        return self._notes_by_uuid[uuid]

    @property
    def initial_seconds(self) -> Optional[float]:
        return self._initial_seconds

    @property
    def uuid(self) -> UUID:
        return self._uuid



from os import PathLike
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Coroutine,
    Dict,
    Optional,
    SupportsFloat,
    SupportsInt,
    Tuple,
    Union,
    runtime_checkable,
)

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

from .enums import AddAction, CalculationRate, HeaderFormat, SampleFormat

if TYPE_CHECKING:
    import numpy

    from .osc import OscBundle, OscMessage


class Default:
    pass


class Missing:
    pass


@runtime_checkable
class SupportsOsc(Protocol):
    def to_osc(self) -> Union["OscBundle", "OscMessage"]:
        ...


@runtime_checkable
class SupportsPlot(Protocol):
    def __plot__(self) -> "numpy.ndarray":
        ...


@runtime_checkable
class SupportsRender(Protocol):
    def __render__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> Coroutine[None, None, Tuple[Optional[Path], int]]:
        ...


@runtime_checkable
class SupportsRenderMemo(Protocol):
    def __render_memo__(self) -> SupportsRender:
        ...


AddActionLike = Optional[Union[AddAction, SupportsInt, str]]
CalculationRateLike = Optional[Union[CalculationRate, SupportsInt, str]]
HeaderFormatLike = Optional[Union[HeaderFormat, SupportsInt, str]]
SampleFormatLike = Optional[Union[SampleFormat, SupportsInt, str]]
UGenInputMap = Optional[Dict[str, Union[SupportsFloat, str, None]]]
